import re
import textwrap

import numpy as np

from pysisyphus.constants import BOHR2ANG, AU2KCALPERMOL
from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.helpers_pure import file_or_str


class MOPAC(Calculator):
    """http://openmopac.net/manual/"""

    conf_key = "mopac"

    MULT_STRS = {
        1: "SINGLET",
        2: "DOUBLET",
        3: "TRIPLET",
        4: "QUARTET",
        5: "QUINTET",
        6: "SEXTET",
        7: "SEPTET",
        8: "OCTET",
    }

    CALC_TYPES = {
        "energy": "1SCF",
        "gradient": "1SCF GRADIENTS",
        "hessian": "DFORCE FORCE LET",
    }

    METHODS = [
        m.lower()
        for m in "AM1 PM3 PM6 PM6-DH2 PM6-D3 PM6-DH+ PM6-DH2 PM6-DH2X "  # lgtm [py/non-iterable-in-for-loop]
        "PM6-D3H4 PM6-D3H4X PM7 PM7-TS".split()
    ]

    def __init__(self, method="PM7", **kwargs):
        super().__init__(**kwargs)

        self.method = method
        assert (
            self.method.lower() in self.METHODS
        ), f"Invalid method={self.method}! Supported methods are ({self.METHODS})"

        self.uhf = "UHF" if self.mult != 1 else ""

        _ = "mopac"
        self.inp_fn = f"{_}.mop"
        self.out_fn = f"{_}.out"
        self.aux_fn = f"{_}.aux"
        self.to_keep = ("mop", "out", "arc", "aux")

        self.parser_funcs = {
            "energy": self.parse_energy,
            "grad": self.parse_grad,
            "hessian": self.parse_hessian,
        }

        self.base_cmd = self.get_cmd()

        """
        1SCF: Do only SCF
        AUX: Creates a checkpoint file
        NOREO: Dont reorient geometry
        """

        self.inp = textwrap.dedent(
            """
        NOSYM {method} {mult} CHARGE={charge} {calc_type} {uhf} THREADS={pal} AUX(6,PRECISION=9) NOREOR

 
        {coord_str}
        """
        ).strip()

        self.log(f"Created MOPAC calculator using the '{self.method}' method.")

    def prepare_coords(self, atoms, coords, opt=False):
        coords = coords.reshape(-1, 3) * BOHR2ANG
        # Optimization flag for coordinate
        of = 1 if opt else 0
        coord_str = "\n".join(
            [
                f"{a} {c[0]: 10.08f} {of} {c[1]: 10.08f} {of} {c[2]: 10.08f} {of}"
                for a, c in zip(atoms, coords)
            ]
        )
        return coord_str

    def prepare_input(self, atoms, coords, calc_type, opt=False):
        coord_str = self.prepare_coords(atoms, coords, opt)

        inp = self.inp.format(
            method=self.method,
            charge=self.charge,
            mult=self.MULT_STRS[self.mult],
            uhf=self.uhf,
            calc_type=self.CALC_TYPES[calc_type],
            coord_str=coord_str,
            pal=self.pal,
            # mem=self.mem,
        )
        return inp

    def get_energy(self, atoms, coords, **prepare_kwargs):
        calc_type = "energy"
        inp = self.prepare_input(atoms, coords, calc_type)
        # with open("inp.mop", "w") as handle:
        # handle.write(inp)
        # import sys; sys.exit()
        results = self.run(inp, calc="energy")
        return results

    def get_forces(self, atoms, coords, **prepare_kwargs):
        calc_type = "gradient"
        inp = self.prepare_input(atoms, coords, calc_type, opt=True)
        results = self.run(inp, calc="grad")
        return results

    def get_hessian(self, atoms, coords, **prepare_kwargs):
        calc_type = "hessian"
        inp = self.prepare_input(atoms, coords, calc_type, opt=True)
        results = self.run(inp, calc="hessian")
        return results

    def run_calculation(self, atoms, coords, **prepare_kwargs):
        return self.get_energy(atoms, coords, **prepare_kwargs)

    def read_aux(self, path):
        with open(path / self.aux_fn) as handle:
            text = handle.read()
        return text

    def parse_energy(self, path):
        return self.parse_energy_from_aux(self.read_aux(path))

    @staticmethod
    @file_or_str(".aux", method=False)
    def parse_energy_from_aux(text):
        energy_re = r"HEAT_OF_FORMATION:KCAL/MOL=([\d\-D+\.]+)"
        mobj = re.search(energy_re, text)
        energy = float(mobj[1].replace("D", "E")) / AU2KCALPERMOL

        result = {
            "energy": energy,
        }
        return result

    def parse_grad(self, path):
        text = self.read_aux(path)
        # grad_re = "GRADIENTS:KCAL.+$\s+(.+)$"
        # mobj = re.search(grad_re, text, re.MULTILINE)
        grad_re = r"GRADIENTS:KCAL/MOL/ANGSTROM\[\d+]=\s+(.+)\s+OVERLAP_MATRIX"
        mobj = re.search(grad_re, text, re.DOTALL)
        # Gradients are given in kcal*mol/angstrom
        gradients = np.array(mobj[1].split(), dtype=float)
        # Convert to hartree/bohr
        gradients = gradients / AU2KCALPERMOL / BOHR2ANG

        forces = -gradients
        result = {
            "forces": forces,
        }
        result.update(self.parse_energy(path))
        return result

    def parse_hessian(self, path):
        return self.parse_hessian_from_aux(self.read_aux(path))

    @staticmethod
    @file_or_str(".aux", method=False)
    def parse_hessian_from_aux(text):
        # Parse employed masses, as the given hessian is mass-weighted
        # and we have to un-weigh it.
        mass_re = re.compile(
            r"ISOTOPIC_MASSES\[(\d+)\]=\s*(.+?)ROTAT_CONSTS", re.DOTALL
        )
        # mobj = re.search(mass_re, text, re.MULTILINE)
        mass_mobj = mass_re.search(text)
        masses = np.array(mass_mobj[2].strip().split(), dtype=float)
        # This matrix is used to un-weigh the hessian
        M = np.diag(np.sqrt(np.repeat(masses, 3)))
        # For N atoms we expect 3N cartesian coordinates
        coord_num = masses.size * 3

        hess_re = r" #  Lower half triangle only\s+([\s\.\-\d]+)\s+NORMAL_MODE"
        tril_hess = re.search(hess_re, text)[1].strip().split()
        tril_hess = np.array(tril_hess, dtype=float)
        assert tril_hess.size == sum(range(coord_num + 1))
        hessian_m = np.zeros((coord_num, coord_num))
        tril_indices = np.tril_indices(coord_num)
        hessian_m[tril_indices] = tril_hess

        triu_indices = np.triu_indices(coord_num, k=1)
        hessian_m[triu_indices] = hessian_m.T[triu_indices]

        # Hessian is given in mdyn/(Å*amu).
        # In a first step we have to unweigh the hessian using the matrix
        # built from the parsed masses.
        hessian = M @ hessian_m @ M
        # Then we have to convert mdyn/Å to Hartree/Bohr²
        #     mdyn/Å = 100 kg/s²
        #     Hartree/Bohr² ~  1556.8931 kg/s²
        #
        #     1 mydn/Å * (100 / 1556.8931 Hartree/Bohr² * Å/mydn) = 0.06423 Hartree/Bohr²
        hessian *= 0.06423

        energy = MOPAC.parse_energy_from_aux(text)["energy"]

        result = {
            "energy": energy,
            "hessian": hessian,
        }
        return result

    def __str__(self):
        return f"MOPAC({self.name})"



import numpy as np
try:
    from xtb.interface import Environment, Param, Calculator as XTBCalculator
    from xtb.libxtb import VERBOSITY_MINIMAL, VERBOSITY_FULL, VERBOSITY_MUTED
except ModuleNotFoundError:
    can_pyxtb = False

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.elem_data import ATOMIC_NUMBERS


class PyXTB(Calculator):

    def __init__(self, *args, gfn=2, acc=None, verbosity=0, keep_calculator=False, **kwargs):
        super().__init__(*args, **kwargs)

        self.env = Environment()
        self.gfn = str(gfn)
        self.acc = acc
        avail_verbosities = {
            0: VERBOSITY_MUTED,
            1: VERBOSITY_MINIMAL,
            2: VERBOSITY_FULL,
        }
        self.verbosity = avail_verbosities[verbosity]
        avail_params = {
            "0": Param.GFN0xTB,
            "1": Param.GFN1xTB,
            "2": Param.GFN2xTB,
            "ff": Param.GFNFF,
        }
        self.keep_calculator = keep_calculator
        self._calculator = None

        self.param = avail_params[self.gfn]
        self.uhf = self.mult - 1

    def get_calculator(self, atoms, coords):
        # Reuse old calculator and updatec coordinates
        if self._calculator:
            self._calculator.update(coords.reshape(-1, 3))
            return self._calculator

        # Create new calculator
        numbers = np.array([ATOMIC_NUMBERS[atom.lower()] for atom in atoms])
        calc = XTBCalculator(
            self.param, numbers, coords.reshape(-1, 3), charge=self.charge, uhf=self.uhf
        )
        calc.set_verbosity(self.verbosity)
        if self.acc:
            calc.set_accuracy(self.acc)

        # Keep calculator, if requested
        if self.keep_calculator and (self._calculator is None):
            self._calculator = calc
        return calc

    def get_energy(self, atoms, coords, **prepare_kwargs):
        calc = self.get_calculator(atoms, coords)
        xtb_result = calc.singlepoint(copy=True)
        results = {
            "energy": xtb_result.get_energy(),
        }
        return results

    def get_forces(self, atoms, coords, **prepare_kwargs):
        calc = self.get_calculator(atoms, coords)
        xtb_result = calc.singlepoint(copy=True)
        results = {
            "energy": xtb_result.get_energy(),
            "forces": -xtb_result.get_gradient().flatten(),
        }
        return results



import logging

from pysisyphus import logger as pysis_logger

logger = pysis_logger.getChild("wavefunction")
logger.setLevel(logging.DEBUG)

from pysisyphus.wavefunction.shells import (
    get_l,
    AOMixShells,
    MoldenShells,
    ORCAShells,
    ORCAMoldenShells,
    Shell,
    Shells,
)

from pysisyphus.wavefunction.excited_states import norm_ci_coeffs
from pysisyphus.wavefunction.wavefunction import Wavefunction
from pysisyphus.wavefunction.localization import (
    cholesky,
    edmiston_ruedenberg,
    foster_boys,
    pipek_mezey,
)



import numpy as np
import rmsd

from pysisyphus.stocastic.Pipeline import Pipeline


class Kick(Pipeline):

    def __init__(self, geom, radius=0.5, **kwargs):
        super().__init__(geom, **kwargs)

        self.radius = radius

    def get_kick(self):
        # Interval [0, 1)
        kick = np.random.random(4*self.coords_size)
        # Stretch [0, 1) to  [-r ... r)
        kick = self.radius * (2*kick - 1)
        kick = kick.reshape(-1, 3)
        # Filter for kicks within a sphere with radius self.radius
        kick_lengths = np.linalg.norm(kick, axis=1)
        valid_kicks = kick[kick_lengths <= self.radius].flatten()
        # Do a recursion when not enough valid kicks were found
        if valid_kicks.size < self.coords_size:
            valid_kicks = self.get_kick()
        # Don't return more than we need
        return valid_kicks[:self.coords_size]

    def get_input_geom(self, geom):
        kick = self.get_kick()
        new_geom = geom.copy()
        new_coords = new_geom.coords + kick
        # Rotate the newly generated coordinates on the initial
        # coordinates.
        # TODO: this may be not needed as we align+match later on ...
        new_coords = rmsd.kabsch_rotate(new_coords.reshape(-1, 3),
                                        self.initial_coords3d
        ).flatten()
        new_geom.coords = new_coords
        return new_geom

    def run_cycle(self, geom):
        print(f"##### Cycle {self.cur_cycle:03d}, "
              f"Micro Cycle {self.cur_micro_cycle:03d} #####")
        opt_geoms = [self.run_kicked_geom(geom) for i in range(self.cycle_size)]
        # Filter out None
        opt_geoms = [geom for geom in opt_geoms if geom]
        opt_num = len(opt_geoms)
        print(f"{opt_num}/{self.cycle_size} optimizations converged.")

        # Comparing to the initial geometry is only useful when the initial
        # geometry is optimized. Otherwise all (small) kicks will converge
        # to the same optimized structure, that is still very different from
        # the initial one.
        """
        initial_rmsds = [
            rmsd.kabsch_rmsd(self.initial_coords3d, ogeom.coords3d)
            for ogeom in opt_geoms
        ]
        print("Initial RMDS")
        print(initial_rmsds)
        """
        kept_geoms = self.get_unique_geometries(opt_geoms)

        # cycle_str = f"{self.cur_cycle:03d}_{self.cur_micro_cycle:03d}"
        # fn_base = f"cycle_{cycle_str}"
        # trj_fn = f"{fn_base}_trj.xyz"
        # with open(trj_fn, "w") as handle:
            # handle.write(make_trj_str_from_geoms(opt_geoms))

        # trj_filtered_fn = f"{fn_base}_filtered_trj.xyz"
        # with open(trj_filtered_fn, "w") as handle:
            # handle.write(make_trj_str_from_geoms(kept_geoms))

        # for i, ogeom in enumerate(kept_geoms):
            # fn = f"geom_{i:02d}_{cycle_str}.xyz"
            # with open(fn, "w") as handle:
                # handle.write(ogeom.as_xyz())
        self.cur_micro_cycle += 1

        return kept_geoms



from math import sqrt
from pathlib import Path
from typing import Literal, Optional

import numpy as np
from scipy.optimize import root_scalar

from pysisyphus.cos.ChainOfStates import ChainOfStates
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers_pure import rms
# from pysisyphus.io.hessian import save_hessian
from pysisyphus.optimizers.guess_hessians import (
    get_guess_hessian,
    xtb_hessian,
    HessInit,
)
from pysisyphus.optimizers.hessian_updates import (
    bfgs_update,
    flowchart_update,
    damped_bfgs_update,
    bofill_update,
    ts_bfgs_update,
    ts_bfgs_update_org,
    ts_bfgs_update_revised,
)
from pysisyphus.optimizers.Optimizer import Optimizer
from pysisyphus.optimizers.exceptions import OptimizationError

from pysisyphus.helpers import array2string
import torch

def dummy_hessian_update(H, dx, dg):
    return np.zeros_like(H), "no"


HESS_UPDATE_FUNCS = {
    "none": dummy_hessian_update,
    None: dummy_hessian_update,
    False: dummy_hessian_update,
    "bfgs": bfgs_update,
    "damped_bfgs": damped_bfgs_update,
    "flowchart": flowchart_update,
    "bofill": bofill_update,
    "ts_bfgs": ts_bfgs_update,
    "ts_bfgs_org": ts_bfgs_update_org,
    "ts_bfgs_rev": ts_bfgs_update_revised,
}
HessUpdate = Literal[
    "none",
    None,
    False,
    "bfgs",
    "damped_bfgs",
    "flowchart",
    "bofill",
    "ts_bfgs",
    "ts_bfgs_org",
    "ts_bfgs_rev",
]


class HessianOptimizer(Optimizer):
    rfo_dict = {
        "min": (0, "min"),
        "max": (-1, "max"),
    }

    def __init__(
        self,
        geometry: Geometry,
        trust_radius: float = 0.5,
        trust_update: bool = True,
        trust_min: float = 0.1,
        trust_max: float = 1,
        max_energy_incr: Optional[float] = None,
        hessian_update: HessUpdate = "bfgs",
        hessian_init: HessInit = "fischer",
        hessian_recalc: Optional[int] = None,
        hessian_recalc_adapt: Optional[float] = None,
        hessian_xtb: bool = False,
        hessian_recalc_reset: bool = False,
        small_eigval_thresh: float = 1e-8,
        line_search: bool = False,
        alpha0: float = 1.0,
        max_micro_cycles: int = 25,
        rfo_overlaps: bool = False,
        **kwargs,
    ) -> None:
        """Baseclass for optimizers utilizing Hessian information.

        Parameters
        ----------
        geometry
            Geometry to be optimized.
        trust_radius
            Initial trust radius in whatever unit the optimization is carried out.
        trust_update
            Whether to update the trust radius throughout the optimization.
        trust_min
            Minimum trust radius.
        trust_max
            Maximum trust radius.
        max_energy_incr
            Maximum allowed energy increased after a faulty step. Optimization is
            aborted when the threshold is exceeded.
        hessian_update
            Type of Hessian update. Defaults to BFGS for minimizations and Bofill
            for saddle point searches.
        hessian_init
            Type of initial model Hessian.
        hessian_recalc
            Recalculate exact Hessian every n-th cycle instead of updating it.
        hessian_recalc_adapt
            Use a more flexible scheme to determine Hessian recalculation. Undocumented.
        hessian_xtb
            Recalculate the Hessian at the GFN2-XTB level of theory.
        hessian_recalc_reset
            Whether to skip Hessian recalculation after reset. Undocumented.
        small_eigval_thresh
            Threshold for small eigenvalues. Eigenvectors belonging to eigenvalues
            below this threshold are discardewd.
        line_search
            Whether to carry out a line search. Not implemented by a subclassing
            optimizers.
        alpha0
            Initial alpha for restricted-step (RS) procedure.
        max_micro_cycles
            Maximum number of RS iterations.
        rfo_overlaps
            Enable mode-following in RS procedure.

        Other Parameters
        ----------------
        **kwargs
            Keyword arguments passed to the Optimizer baseclass.
        """
        super().__init__(geometry, **kwargs)

        assert not issubclass(
            type(geometry), ChainOfStates
        ), "HessianOptimizer can't be used for and ChainOfStates objects!"

        self.trust_update = bool(trust_update)
        assert trust_min <= trust_max, "trust_min must be <= trust_max!"
        self.trust_min = float(trust_min)
        self.trust_max = float(trust_max)
        self.max_energy_incr = max_energy_incr
        # Constrain initial trust radius if trust_max > trust_radius
        self.trust_radius = min(trust_radius, trust_max)
        self.log(f"Initial trust radius: {self.trust_radius:.6f}")
        self.hessian_update = hessian_update
        self.hessian_update_func = HESS_UPDATE_FUNCS[hessian_update]
        self.hessian_init = hessian_init
        self.hessian_recalc = hessian_recalc
        self.hessian_recalc_adapt = hessian_recalc_adapt
        self.hessian_xtb = hessian_xtb
        self.hessian_recalc_reset = hessian_recalc_reset
        self.small_eigval_thresh = float(small_eigval_thresh)
        self.line_search = bool(line_search)
        # Restricted-step related
        self.alpha0 = alpha0
        self.max_micro_cycles = int(max_micro_cycles)
        assert max_micro_cycles >= 0
        self.rfo_overlaps = rfo_overlaps

        assert self.small_eigval_thresh > 0.0, "small_eigval_thresh must be > 0.!"
        if not self.restarted:
            self.hessian_recalc_in = None
            self.adapt_norm = None
            self.predicted_energy_changes = list()
        hessian_init_exists = Path(self.hessian_init).exists()
        if (
            # Allow actually calculated Hessians for all coordinate systems
            not hessian_init_exists
            and self.hessian_init not in ("calc", "xtb", "xtb1", "xtbff")
            # But disable model Hessian for Cartesian optimizations
            and self.geometry.coord_type in ("cart", "cartesian", "mwcartesian")
        ):
            self.hessian_init = "unit"
            self.log(
                f"Chosen initial (model) Hessian is incompatible with current "
                f"coord_type: {self.geometry.coord_type}!"
            )

        self._prev_eigvec_min = None
        self._prev_eigvec_max = None
        self._using_active_dofs = False
        self._active_dof_indices = None
        self.cur_H = None

    @property
    def using_active_dofs(self):
        return self._using_active_dofs

    @property
    def active_dof_indices(self):
        return self._active_dof_indices

    def _set_active_dofs(self, use_active):
        self._using_active_dofs = use_active
        if not use_active:
            self._active_dof_indices = None
            return
        if getattr(self.geometry, "within_partial_hessian", None) is not None:
            self._active_dof_indices = self.geometry.hess_active_dof_indices
            return
        # Fallback: infer active DOFs from calculator's Hessian-active atoms
        calc = getattr(self.geometry, "calculator", None)
        core = getattr(calc, "core", calc)
        hess_atoms = getattr(core, "hess_active_atoms", None)
        if hess_atoms is not None and len(hess_atoms) > 0:
            act = []
            for a in hess_atoms:
                base = 3 * int(a)
                act.extend([base, base + 1, base + 2])
            self._active_dof_indices = np.asarray(act, dtype=int)
            return
        self._active_dof_indices = self.geometry.active_dof_indices

    def active_from_full(self, vec):
        if not self.using_active_dofs:
            return vec
        inds = self._active_dof_indices
        if inds is None:
            return vec
        # Sanitize indices (drop negatives / out-of-bounds)
        try:
            inds = np.asarray(inds, dtype=int)
            if len(inds) > 0:
                if np.any(inds < 0):
                    inds = inds[inds >= 0]
                if len(inds) > 0:
                    max_valid = vec.shape[0] - 1
                    if np.any(inds > max_valid):
                        inds = inds[inds <= max_valid]
                if len(inds) > 0:
                    if np.min(inds) < 0 or np.max(inds) >= vec.shape[0]:
                        return vec
        except (ValueError, IndexError, TypeError):
            pass
        # Avoid double-slicing if vector is already in active space
        try:
            if vec.shape[0] == len(inds):
                return vec
        except (ValueError, IndexError, TypeError):
            pass
        try:
            if len(inds) > 0 and vec.shape[0] <= int(np.max(inds)):
                # Indices exceed vector length → already active (compact) vector.
                return vec
        except (ValueError, IndexError, TypeError):
            pass
        if isinstance(vec, torch.Tensor):
            if vec.device.type == "cuda":
                try:
                    vec_cpu = vec.detach().cpu().numpy()
                    return torch.as_tensor(vec_cpu[inds], dtype=vec.dtype, device=vec.device)
                except (ValueError, IndexError, TypeError):
                    return vec
            idx = torch.as_tensor(inds, dtype=torch.long, device=vec.device)
            return vec.index_select(0, idx)
        return vec[inds]

    def full_from_active(self, vec):
        if not self.using_active_dofs:
            return vec
        inds = self._active_dof_indices
        if inds is None:
            return vec
        if isinstance(vec, torch.Tensor):
            idx = torch.as_tensor(inds, dtype=torch.long, device=vec.device)
            full = torch.zeros(self.geometry.cart_coords.size, dtype=vec.dtype, device=vec.device)
            full.index_copy_(0, idx, vec)
            return full
        full = np.zeros(self.geometry.cart_coords.size, dtype=vec.dtype if hasattr(vec, "dtype") else float)
        full[inds] = vec
        return full

    def active_hessian(self, hessian):
        if not self.using_active_dofs:
            return hessian

        if getattr(self.geometry, "within_partial_hessian", None) is not None:
            act_n_dof = int(self.geometry.within_partial_hessian.get("active_n_dof", 0))
            if hessian.shape == (act_n_dof, act_n_dof):
                return hessian

        inds = self.active_dof_indices
        try:
            inds_arr = np.asarray(inds, dtype=int)
            if hessian.shape[0] == len(inds_arr) and len(inds_arr) > 0:
                if np.max(inds_arr) >= hessian.shape[0]:
                    # Likely already in active order; avoid double-slicing.
                    return hessian
        except (ValueError, IndexError, TypeError):
            pass
        try:
            inds = np.asarray(inds, dtype=int)
            if len(inds) > 0:
                max_valid = hessian.shape[0] - 1
                inds = inds[(inds >= 0) & (inds <= max_valid)]
        except (ValueError, IndexError, TypeError):
            pass
        if isinstance(hessian, torch.Tensor):
            if hessian.device.type == "cuda":
                try:
                    hess_cpu = hessian.detach().cpu().numpy()
                    return torch.as_tensor(hess_cpu[np.ix_(inds, inds)], dtype=hessian.dtype, device=hessian.device)
                except (ValueError, IndexError, TypeError):
                    return hessian
            idx = torch.as_tensor(inds, device=hessian.device, dtype=torch.int64)
            return hessian.index_select(0, idx).index_select(1, idx)
        return hessian[np.ix_(inds, inds)]

    def active_list(self, seq):
        if not self.using_active_dofs:
            return seq
        return [self.active_from_full(item) for item in seq]

    @property
    def prev_eigvec_min(self):
        return self._prev_eigvec_min

    @prev_eigvec_min.setter
    def prev_eigvec_min(self, prev_eigvec_min):
        if self.rfo_overlaps:
            self._prev_eigvec_min = prev_eigvec_min

    @property
    def prev_eigvec_max(self):
        return self._prev_eigvec_max

    @prev_eigvec_max.setter
    def prev_eigvec_max(self, prev_eigvec_max):
        if self.rfo_overlaps:
            self._prev_eigvec_max = prev_eigvec_max

    def reset(self):
        # Don't recalculate the hessian if we have to reset the optimizer
        hessian_init = self.hessian_init
        if (
            (not self.hessian_recalc_reset)
            and hessian_init == "calc"
            and self.geometry.coord_type != "cart"
        ):
            hessian_init = "fischer"
        self.prepare_opt(hessian_init)

    # def save_hessian(self):
    #     # Don't try to save Hessians of analytical potentials
    #     if self.geometry.is_analytical_2d:
    #         return

    #     h5_fn = self.get_path_for_fn(f"hess_calc_cyc_{self.cur_cycle}.h5")
    #     # Save the cartesian hessian, as it is independent of the
    #     # actual coordinate system that is used.
    #     save_hessian(
    #         h5_fn,
    #         self.geometry,
    #         self.geometry.cart_hessian,
    #         self.geometry.energy,
    #         self.geometry.calculator.mult,
    #     )
    #     self.log(f"Wrote calculated cartesian Hessian to '{h5_fn}'")

    def prepare_opt(self, hessian_init=None):
        if hessian_init is None:
            hessian_init = self.hessian_init

        self.H, hess_str = get_guess_hessian(self.geometry, hessian_init)
        if self.hessian_init != "calc" and self.geometry.is_analytical_2d:
            assert self.H.shape == (3, 3)
            self.H[2, 2] = 0.0

        msg = f"Using {hess_str} Hessian"
        if hess_str == "saved":
            msg += f" from '{hessian_init}'"
        self.log(msg)

        # # Dump to disk if hessian was calculated
        # if self.hessian_init == "calc":
        #     self.save_hessian()

        if (
            hasattr(self.geometry, "coord_type")
            and self.geometry.coord_type == "dlc"
            # Calculated Hessian is already in DLC
            and hessian_init != "calc"
        ):
            U = self.geometry.internal.U
            self.H = U.T.dot(self.H).dot(U)

        if self.hessian_recalc_adapt:
            self.adapt_norm = np.linalg.norm(self.geometry.forces)

        if self.hessian_recalc:
            # Already substract one, as we don't do a hessian update in
            # the first cycle.
            self.hessian_recalc_in = self.hessian_recalc - 1

    def _get_opt_restart_info(self):
        opt_restart_info = {
            "adapt_norm": self.adapt_norm,
            "H": self.H.tolist(),
            "hessian_recalc_in": self.hessian_recalc_in,
            "predicted_energy_changes": self.predicted_energy_changes,
        }
        return opt_restart_info

    def _set_opt_restart_info(self, opt_restart_info):
        self.adapt_norm = opt_restart_info["adapt_norm"]
        self.H = np.array(opt_restart_info["H"])
        self.hessian_recalc_in = opt_restart_info["hessian_recalc_in"]
        self.predicted_energy_changes = opt_restart_info["predicted_energy_changes"]

    def update_trust_radius(self):
        # The predicted change should be calculated at the end of optimize
        # of the previous cycle.
        assert (
            len(self.predicted_energy_changes) == len(self.forces) - 1
        ), "Did you forget to append to self.predicted_energy_changes?"
        self.log("Trust radius update")
        self.log(f"\tCurrent trust radius: {self.trust_radius:.6f}")
        predicted_change = self.predicted_energy_changes[-1]
        actual_change = self.energies[-1] - self.energies[-2]
        # Only report an unexpected increase if we actually predicted a
        # decrease.
        unexpected_increase = (actual_change > 0) and (predicted_change < 0)
        old_trust = self.trust_radius
        if unexpected_increase:
            self.log(f"Energy increased by {actual_change:.6f} au!")
            if self.max_energy_incr and (actual_change > self.max_energy_incr):
                raise OptimizationError("Actual energy change too high!")
        coeff = actual_change / predicted_change
        self.log(f"\tPredicted change: {predicted_change:.4e} au")
        self.log(f"\tActual change: {actual_change:.4e} au")
        self.log(f"\tCoefficient: {coeff:.2%}")
        step = self.steps[-1]
        last_step_norm = np.linalg.norm(step)
        self.set_new_trust_radius(coeff, last_step_norm)
        if unexpected_increase:
            self.table.print(
                f"Unexpected energy increase ({actual_change:.6f} au)! "
                f"Trust radius: old={old_trust:.4}, new={self.trust_radius:.4}"
            )

    def set_new_trust_radius(self, coeff, last_step_norm):
        # Nocedal, Numerical optimization Chapter 4, Algorithm 4.1

        # If actual and predicted energy change have different signs
        # coeff will be negative and lead to a decreased trust radius,
        # which is fine.
        if coeff < 0.25:
            self.trust_radius = max(self.trust_radius / 4, self.trust_min)
            self.log("\tDecreasing trust radius.")
        # Only increase trust radius if last step norm was at least 80% of it
        # See [5], Appendix, step size and direction control
        # elif coeff > 0.75 and (last_step_norm >= .8*self.trust_radius):
        #
        # Only increase trust radius if last step norm corresponded approximately
        # to the trust radius.
        elif coeff > 0.75 and abs(self.trust_radius - last_step_norm) <= 1e-3:
            self.trust_radius = min(self.trust_radius * 2, self.trust_max)
            self.log("\tIncreasing trust radius.")
        else:
            self.log(f"\tKeeping current trust radius at {self.trust_radius:.6f}")
            return
        self.log(f"\tUpdated trust radius: {self.trust_radius:.6f}")

    def update_hessian(self):
        # Compare current forces to reference forces to see if we shall recalc the
        # hessian.
        try:
            cur_norm = np.linalg.norm(self.forces[-1])
            ref_norm = self.adapt_norm / self.hessian_recalc_adapt
            recalc_adapt = cur_norm <= ref_norm
            self.log(
                "Check for adaptive Hessian recalculation: "
                f"{cur_norm:.6f} <= {ref_norm:.6f}, {recalc_adapt}"
            )
        except TypeError:
            recalc_adapt = False

        try:
            self.hessian_recalc_in = max(self.hessian_recalc_in - 1, 0)
            self.log(f"Recalculation of Hessian in {self.hessian_recalc_in} cycle(s).")
        except TypeError:
            self.hessian_recalc_in = None

        # Update reference norm if needed
        # TODO: Decide on whether to update the norm when the recalculation is
        # initiated by 'recalc'.
        if recalc_adapt:
            self.adapt_norm = cur_norm

        recalc = self.hessian_recalc_in == 0

        if recalc or recalc_adapt:
            # Free old Hessian from GPU before recalculating
            H_old = self.H
            self.H = None
            del H_old
            try:
                import torch
                if torch.cuda.is_available():
                    torch.cuda.empty_cache()
            except ImportError:
                pass
            # Use xtb hessian
            self.log("Requested Hessian recalculation.")
            if self.hessian_xtb:
                self.H = xtb_hessian(self.geometry)
                key = "xtb"
            # Calculated hessian at actual level of theory
            else:
                self.H = self.geometry.hessian
                key = "exact"
                # self.save_hessian()
            if self.using_active_dofs:
                # Keep the optimizer Hessian in active-DOF space to avoid
                # shape mismatches during quasi-Newton updates.
                self.H = self.active_hessian(self.H)
            if not (self.cur_cycle == 0):
                self.log(f"Recalculated {key} Hessian in cycle {self.cur_cycle}.")
            # Reset counter. It is also reset when the recalculation was initiated
            # by the adaptive formulation.
            self.hessian_recalc_in = self.hessian_recalc
        # Simple hessian update
        else:
            dx = self.steps[-1]
            dg = -(self.forces[-1] - self.forces[-2])
            H_work = self.H
            if self.using_active_dofs:
                H_work = self.active_hessian(self.H)
                dx = self.active_from_full(dx)
                dg = self.active_from_full(dg)
            curv_cond = dx.dot(dg)
            if curv_cond < 0.0:
                self.log(
                    f"Curvature condition (s·y = {curv_cond:.4f} < 0) not satisfied!"
                )
            dH, key = self.hessian_update_func(H_work, dx, dg)
            self.H = H_work + dH
            self.log(f"Did {key} Hessian update.")

    def solve_rfo(self, rfo_mat, kind="min", prev_eigvec=None):
        # When using the restricted step variant of RFO the RFO matrix
        # may not be symmetric. Thats why we can't use eigh here.
        is_torch = isinstance(rfo_mat, torch.Tensor)

        if is_torch:
            if not torch.isfinite(rfo_mat).all():
                self.log("RFO matrix contains NaN/inf; sanitizing entries.")
                rfo_mat = torch.nan_to_num(
                    rfo_mat, nan=0.0, posinf=1e8, neginf=-1e8
                )
        else:
            if not np.isfinite(rfo_mat).all():
                self.log("RFO matrix contains NaN/inf; sanitizing entries.")
                rfo_mat = np.nan_to_num(rfo_mat, nan=0.0, posinf=1e8, neginf=-1e8)

        if is_torch:
            is_sym = torch.allclose(rfo_mat, rfo_mat.T)
        else:
            is_sym = np.allclose(rfo_mat, rfo_mat.T)

        if is_sym:
            try:
                eigenvalues, eigenvectors = (torch.linalg.eigh(rfo_mat) if is_torch else np.linalg.eigh(rfo_mat))
            except (torch._C._LinAlgError, np.linalg.LinAlgError):
                self.log("eigh failed; falling back to eig.")
                eigenvalues, eigenvectors = (torch.linalg.eig(rfo_mat) if is_torch else np.linalg.eig(rfo_mat))
                eigenvalues = eigenvalues.real
                eigenvectors = eigenvectors.real
        else:
            eigenvalues, eigenvectors = (torch.linalg.eig(rfo_mat) if is_torch else np.linalg.eig(rfo_mat))
            eigenvalues = eigenvalues.real
            eigenvectors = eigenvectors.real

        self.log("\tdiagonalized augmented Hessian")

        if isinstance(eigenvectors, torch.Tensor):
            sorted_inds = torch.argsort(eigenvalues)
        else:
            sorted_inds = np.argsort(eigenvalues)

        # Depending on wether we want to minimize (maximize) along
        # the mode(s) in the rfo mat we have to select the smallest
        # (biggest) eigenvalue and corresponding eigenvector.
        first_or_last, verbose = self.rfo_dict[kind]
        # Given sorted eigenvalue-indices (sorted_inds) use the first
        # (smallest eigenvalue) or the last (largest eigenvalue) index.
        if prev_eigvec is None:
            ind = sorted_inds[first_or_last]
        else:
            if isinstance(prev_eigvec, torch.Tensor):
                ovlps = prev_eigvec.matmul(eigenvectors)
            else:
                ovlps = np.array([prev_eigvec.dot(ev) for ev in eigenvectors.T])
            naive_ind = sorted_inds[first_or_last]
            ind = np.abs(ovlps).argmax() if isinstance(ovlps, np.ndarray) else torch.argmax(torch.abs(ovlps)).item()
            self.log(
                f"Overlap: {ind} ({eigenvalues[ind]:.6f}), "
                f"Naive: {naive_ind} ({eigenvalues[naive_ind]:.6f})"
            )
        follow_eigvec = eigenvectors.T[ind]
        if isinstance(follow_eigvec, torch.Tensor):
            step_nu = follow_eigvec.clone()
        else:
            step_nu = follow_eigvec.copy()
        nu = step_nu[-1]
        self.log(f"\tnu_{verbose}={nu:.8e}")
        # Scale eigenvector so that its last element equals 1. The
        # final is step is the scaled eigenvector without the last element.
        step = step_nu[:-1] / nu
        eigval = eigenvalues[ind]
        self.log(f"\teigenvalue_{verbose}={eigval:.8e}")
        return step, eigval, nu, follow_eigvec

    def solve_rfo_secular(self, eigvals, gradient, alpha=1.0, kind="min",
                          prev_eigvec=None, max_iter=50, tol=1e-12):
        """Solve the RFO eigenvalue problem via the secular equation.

        The augmented Hessian has arrowhead structure, so its eigenvalue
        problem reduces to: f(mu) = sum g_i^2/(alpha*mu - lam_i) - mu = 0,
        solvable in O(N) instead of O(N^3).

        Returns (step, eigval, nu, eigvec) on success, None on failure.
        """
        is_torch = isinstance(eigvals, torch.Tensor)

        # Convert all inputs to plain numpy/float for root-finding
        _np = lambda x: x.detach().cpu().numpy().copy() if isinstance(x, torch.Tensor) else np.asarray(x, dtype=np.float64)
        g = _np(gradient)
        lam = _np(eigvals)
        alpha = alpha.detach().cpu().item() if isinstance(alpha, torch.Tensor) else float(alpha)

        n = len(lam)
        g2 = g ** 2
        nz = g2 > 1e-30
        if not nz.any():
            step = np.zeros(n)
            eigvec = np.zeros(n + 1); eigvec[-1] = 1.0
            if is_torch:
                step = torch.zeros(n, device=eigvals.device, dtype=eigvals.dtype)
                eigvec = torch.zeros(n + 1, device=eigvals.device, dtype=eigvals.dtype)
                eigvec[-1] = 1.0
            return step, 0.0, 1.0, eigvec

        g2_nz, lam_nz = g2[nz], lam[nz]

        # Guard against alpha ≈ 0 (can arise from trust-radius adaptation)
        if abs(alpha) < 1e-14:
            return None

        def f_df(mu):
            d = alpha * mu - lam_nz
            return float(np.sum(g2_nz / d) - mu), float(-alpha * np.sum(g2_nz / d**2) - 1.0)

        _, verbose = self.rfo_dict[kind]

        # Bracket the root
        if kind == "min":
            pole = float(lam.min() / alpha)
            mu = pole - max(float(np.sqrt(g2_nz.sum())) / alpha, 1.0)
            for _ in range(20):
                if f_df(mu)[0] > 0: break
                mu = pole - 2.0 * (pole - mu)
            else:
                return None
            lo, hi = mu, pole - 1e-15 * max(abs(pole), 1.0)
        elif kind == "max":
            pole = float(lam.max() / alpha)
            mu = pole + max(float(np.sqrt(g2_nz.sum())) / alpha, 1.0)
            for _ in range(20):
                if f_df(mu)[0] < 0: break
                mu = pole + 2.0 * (mu - pole)
            else:
                return None
            lo, hi = pole + 1e-15 * max(abs(pole), 1.0), mu
        else:
            return None

        # Newton-Raphson with bisection safeguard
        mu_cur = (lo + hi) / 2.0
        for _ in range(max_iter):
            fval, dfval = f_df(mu_cur)
            if abs(fval) < tol:
                break
            mu_new = mu_cur - fval / dfval if abs(dfval) > 1e-30 else (lo + hi) / 2.0
            if mu_new <= lo or mu_new >= hi:
                mu_new = (lo + hi) / 2.0
            f_new = f_df(mu_new)[0]
            if f_new > 0: lo = mu_new
            else: hi = mu_new
            mu_cur = mu_new
        else:
            self.log(f"Secular equation did not converge in {max_iter} iters.")
            return None

        self.log(f"\teigenvalue_{verbose}={mu_cur:.8e} (secular)")
        self.log(f"\tnu_{verbose}={1.0:.8e}")

        # Compute step: s_i = g_i / (alpha * mu - lam_i)
        denom = alpha * mu_cur - lam
        step_np = np.where(nz, g / denom, 0.0)

        # Eigenvector for mode tracking
        eigvec_np = np.append(step_np, 1.0)
        eigvec_np /= np.linalg.norm(eigvec_np)

        # Mode tracking check
        if prev_eigvec is not None:
            prev_np = _np(prev_eigvec)
            if abs(float(np.dot(prev_np, eigvec_np))) < 0.3:
                self.log("Secular eigvec overlap too low; falling back.")
                return None

        if is_torch:
            step_np = torch.tensor(step_np, device=eigvals.device, dtype=eigvals.dtype)
            eigvec_np = torch.tensor(eigvec_np, device=eigvals.device, dtype=eigvals.dtype)

        return step_np, mu_cur, 1.0, eigvec_np

    def filter_small_eigvals(self, eigvals, eigvecs, mask=False):
        if isinstance(eigvals, torch.Tensor):
            small_inds = torch.abs(eigvals) < self.small_eigval_thresh
        else:
            small_inds = np.abs(eigvals) < self.small_eigval_thresh
        eigvals = eigvals[~small_inds]
        eigvecs = eigvecs[:, ~small_inds]
        small_num = sum(small_inds)
        self.log(
            f"Found {small_num} small eigenvalues in Hessian. Removed "
            "corresponding eigenvalues and eigenvectors."
        )
        # assert small_num <= 6, (
        #     "Expected at most 6 small eigenvalues in cartesian hessian "
        #     f"but found {small_num}!"
        # )
        if mask:
            return eigvals, eigvecs, small_inds
        else:
            return eigvals, eigvecs

    def log_negative_eigenvalues(self, eigvals, pre_str=""):
        neg_inds = eigvals < -self.small_eigval_thresh
        neg_eigval_str = array2string(eigvals[neg_inds], precision=6)
        self.log(f"{pre_str}Hessian has {neg_inds.sum()} negative eigenvalue(s).")
        self.log(f"\t{neg_eigval_str}")

    def housekeeping(self):
        """Calculate gradient and energy. Update trust radius and hessian
        if needed. Return energy, gradient and hessian for the current cycle."""
        gradient_full = self.geometry.gradient
        energy = self.geometry.energy
        self.energies.append(energy)
        self.log(f"    Energy: {energy: >12.6f} au")
        self.log(
            f"norm(grad): {np.linalg.norm(gradient_full): >12.6f} au / bohr (rad)"
        )
        self.log(
            f" rms(grad): {np.sqrt(np.mean(gradient_full**2)): >12.6f} au / bohr (rad)"
        )
        self.forces.append(-gradient_full)

        can_update = (
            # Allows gradient differences
            len(self.forces) > 1
            and (self.forces[-2].shape == gradient_full.shape)
            and len(self.coords) > 1
            # Coordinates may have been rebuilt. Take care of that.
            and (self.coords[-2].shape == self.coords[1].shape)
            and len(self.energies) > 1
        )
        if can_update:
            if self.trust_update:
                self.update_trust_radius()
            self.update_hessian()

        # Convert gradient to match H device/dtype AFTER update_hessian(),
        # so that hessian_recalc (which may replace self.H with a new tensor
        # on a different device) is accounted for.
        if isinstance(self.H, torch.Tensor):
            gradient_full = torch.from_numpy(gradient_full).to(
                self.H.device, self.H.dtype
            )

        H = self.H
        if self.geometry.internal:
            # Shift eigenvalues of orthogonal part to high values, so they
            # don't contribute to the actual step.
            H_proj = self.geometry.internal.project_hessian(self.H)
            # Symmetrize hessian, as the projection may break it?!
            H = (H_proj + H_proj.T) / 2

        if getattr(self.geometry, "within_partial_hessian", None) is not None:
            use_active = True
        elif (
            H.shape[0] != self.geometry.cart_coords.size
            and self.geometry.coord_type in ("cart", "cartesian", "mwcartesian")
        ):
            # Partial Hessian without explicit metadata: still use active slicing.
            # Only applies to Cartesian coordinate types; for internal coordinates
            # (e.g. DLC), the Hessian is naturally smaller than cart_coords.size.
            use_active = True
        else:
            use_active = (
                len(self.geometry.freeze_atoms) > 0
                and self.geometry.coord_type in ("cart", "cartesian", "mwcartesian")
                and H.shape[0] == self.geometry.cart_coords.size
            )
        self._set_active_dofs(use_active)

        H = self.active_hessian(H)
        if gradient_full.shape[0] == H.shape[0]:
            gradient = gradient_full
        else:
            gradient = self.active_from_full(gradient_full)

        if isinstance(H, torch.Tensor):
            eigvals, eigvecs = torch.linalg.eigh(H)
        else:
            eigvals, eigvecs = np.linalg.eigh(H)
        # Neglect small eigenvalues
        eigvals, eigvecs = self.filter_small_eigvals(eigvals, eigvecs)

        resetted = not can_update
        self.cur_H = H
        return energy, gradient, H, eigvals, eigvecs, resetted

    def get_augmented_hessian(self, eigvals, gradient, alpha=1.0):
        if isinstance(gradient, torch.Tensor):
            dim_ = eigvals.size(0) + 1
            H_aug = torch.zeros((dim_, dim_), device=gradient.device, dtype=gradient.dtype)
            H_aug[: dim_ - 1, : dim_ - 1] = torch.diag(eigvals / alpha)
        else:
            dim_ = eigvals.size + 1
            H_aug = np.zeros((dim_, dim_))
            H_aug[: dim_ - 1, : dim_ - 1] = np.diag(eigvals / alpha)
        H_aug[-1, :-1] = gradient
        H_aug[:-1, -1] = gradient

        H_aug[:-1, -1] /= alpha

        return H_aug

    def get_alpha_step(self, cur_alpha, rfo_eigval, step_norm, eigvals, gradient):
        # Derivative of the squared step w.r.t. alpha
        numer = gradient**2
        denom = (eigvals - rfo_eigval * cur_alpha) ** 3
        if isinstance(gradient, torch.Tensor):
            quot = torch.sum(numer / denom)
        else:
            quot = np.sum(numer / denom)
        self.log(f"quot={quot:.6f}")
        dstep2_dalpha = 2 * rfo_eigval / (1 + step_norm**2 * cur_alpha) * quot
        if isinstance(gradient, torch.Tensor):
            dstep2_valid = bool(
                torch.isfinite(dstep2_dalpha)
                & (torch.abs(dstep2_dalpha) > 1e-12)
            )
        else:
            dstep2_valid = np.isfinite(dstep2_dalpha) and abs(dstep2_dalpha) > 1e-12
        if not dstep2_valid:
            self.log(
                "alpha update skipped due to invalid derivative "
                f"(dstep2_dalpha={dstep2_dalpha})"
            )
            return 0.0
        self.log(f"analytic deriv.={dstep2_dalpha:.6f}")
        # Update alpha
        alpha_step = (
            2 * (self.trust_radius * step_norm - step_norm**2) / dstep2_dalpha
        )
        self.log(f"alpha_step={alpha_step:.4f}")
        min_alpha = 1e-8
        if (cur_alpha + alpha_step) <= min_alpha:
            self.log(
                "alpha update would make alpha non-positive; "
                f"clamping to min_alpha={min_alpha:.1e}"
            )
            alpha_step = min_alpha - cur_alpha
        return alpha_step

    def get_rs_step(self, eigvals, eigvecs, gradient, name="RS"):
        # Transform gradient to basis of eigenvectors
        if isinstance(eigvecs, torch.Tensor):
            if not isinstance(gradient, torch.Tensor):
                gradient = torch.as_tensor(
                    gradient, device=eigvecs.device, dtype=eigvecs.dtype
                )
            elif gradient.device != eigvecs.device:
                gradient = gradient.to(device=eigvecs.device)
            gradient_ = eigvecs.T @ gradient
        else:
            gradient_ = eigvecs.T.dot(gradient)

        alpha = self.alpha0
        for mu in range(self.max_micro_cycles):
            self.log(f"{name} micro cycle {mu:02d}, alpha={alpha:.6f}")
            # Try secular equation solver first (O(N) vs O(N^3))
            secular_result = self.solve_rfo_secular(
                eigvals, gradient_, alpha, kind="min",
                prev_eigvec=self.prev_eigvec_min,
            )
            if secular_result is not None:
                rfo_step_, eigval_min, nu, self.prev_eigvec_min = secular_result
            else:
                # Fallback to full eigendecomposition
                self.log("Secular solver failed; using full eigendecomposition.")
                H_aug = self.get_augmented_hessian(eigvals, gradient_, alpha)
                rfo_step_, eigval_min, nu, self.prev_eigvec_min = self.solve_rfo(
                    H_aug, "min", prev_eigvec=self.prev_eigvec_min
                )
            if isinstance(rfo_step_, torch.Tensor):
                rfo_norm_ = torch.linalg.norm(rfo_step_)
            else:
                rfo_norm_ = np.linalg.norm(rfo_step_)
            self.log(f"norm(rfo step)={rfo_norm_:.6f}")

            if rfo_norm_ <= 0:
                self.log(
                    "RFO step length is zero; falling back to trust-region Newton step."
                )
                step_ = self.get_newton_step_on_trust(
                    eigvals, eigvecs, gradient, transform=False
                )
                break

            if (rfo_norm_ < self.trust_radius) or abs(
                rfo_norm_ - self.trust_radius
            ) <= 1e-3:
                step_ = rfo_step_
                break

            alpha_step = self.get_alpha_step(
                alpha, eigval_min, rfo_norm_, eigvals, gradient_
            )
            alpha += alpha_step
            self.log("")
        # Otherwise, use trust region newton step
        else:
            self.log(
                "RS algorithm did not produce a desired step length in "
                f"{self.max_micro_cycles} micro cycles. Trying RFO with α=1.0."
            )
            secular_result = self.solve_rfo_secular(
                eigvals, gradient_, alpha=1.0, kind="min"
            )
            if secular_result is not None:
                rfo_step_, eigval_min, nu, _ = secular_result
            else:
                H_aug = self.get_augmented_hessian(eigvals, gradient_, alpha=1.0)
                rfo_step_, eigval_min, nu, _ = self.solve_rfo(H_aug, "min")
            if isinstance(rfo_step_, torch.Tensor):
                rfo_norm_ = torch.linalg.norm(rfo_step_)
            else:
                rfo_norm_ = np.linalg.norm(rfo_step_)

            # This should always be True if the above algorithm failed but we
            # keep this line nonetheless,  to make it more obvious.
            if rfo_norm_ > self.trust_radius:
                self.log(
                    f"Proposed RFO step with norm {rfo_norm_:.4f} is outside trust "
                    f"radius Δ={self.trust_radius:.4f}. "
                )
                step_ = self.get_newton_step_on_trust(
                    eigvals, eigvecs, gradient, transform=False
                )
                # Simple, downscaled RFO step
                # step_ = rfo_step_ / rfo_norm_ * self.trust_radius
            else:
                step_ = rfo_step_

        # Transform step back to original basis
        if isinstance(eigvecs, torch.Tensor):
            if isinstance(step_, torch.Tensor):
                pass
            else:
                step_ = torch.tensor(step_, device=eigvecs.device, dtype=eigvecs.dtype)
            step = eigvecs @ step_
            step = step.cpu().numpy()
        else:
            step = eigvecs.dot(step_)
        min_step_norm = getattr(self, "min_step_norm", 0.0)
        step_norm = np.linalg.norm(step)
        if step_norm <= min_step_norm:
            self.log(
                "RFO step length below minimum threshold; "
                "falling back to trust-region Newton step."
            )
            step = self.get_newton_step_on_trust(eigvals, eigvecs, gradient)
        if not np.isfinite(step).all():
            self.log(
                "RFO step contains NaN/inf; falling back to trust-region Newton step."
            )
            step = self.get_newton_step_on_trust(eigvals, eigvecs, gradient)
            if not np.isfinite(step).all():
                raise ValueError(
                    "Fallback Newton step still contains NaN/inf; "
                    "aborting to avoid corrupting coordinates."
                )
        return step

    @staticmethod
    def get_shifted_step_trans(eigvals, gradient_trans, shift):
        return -gradient_trans / (eigvals + shift)

    @staticmethod
    def get_newton_step(eigvals, eigvecs, gradient):
        if isinstance(eigvecs, torch.Tensor):
            eigvals = eigvals.to(eigvecs.device, dtype=eigvecs.dtype)
            gradient = gradient.to(eigvecs.device, dtype=eigvecs.dtype)
            return (eigvecs @ (eigvecs.T @ gradient / eigvals)).cpu().numpy()
        else:
            return eigvecs.dot(eigvecs.T.dot(gradient) / eigvals)

    def get_newton_step_on_trust(self, eigvals, eigvecs, gradient, transform=True):
        """Step on trust-radius.

        See Nocedal 4.3 Iterative solutions of the subproblem
        """
        if isinstance(eigvals, torch.Tensor):
            eigvals = eigvals.cpu().numpy()

        min_ind = eigvals.argmin()
        min_eigval = eigvals[min_ind]
        pos_definite = bool((eigvals > 0.0).all())
        if isinstance(eigvecs, torch.Tensor):
            if not isinstance(gradient, torch.Tensor):
                gradient = torch.tensor(
                    gradient, device=eigvecs.device, dtype=eigvecs.dtype
                )
            else:
                gradient = gradient.to(device=eigvecs.device, dtype=eigvecs.dtype)
            gradient_trans = eigvecs.T @ gradient
            gradient_trans = gradient_trans.cpu().numpy()
        else:
            gradient_trans = eigvecs.T.dot(gradient)

        # This will be also be True when we come close to a minimizer,
        # but then the Hessian will also be positive definite and a
        # simple Newton step will be used.
        hard_case = abs(gradient_trans[min_ind]) <= 1e-6
        self.log(f"Smallest eigenvalue: {min_eigval:.6f}")
        self.log(f"Positive definite Hessian: {pos_definite}")
        self.log(f"Hard case: {hard_case}")

        def get_step(shift):
            return -gradient_trans / (eigvals + shift)

        # Unshifted Newton step
        newton_step_trans = get_step(0.0)
        newton_norm = np.linalg.norm(newton_step_trans)

        def on_trust_radius_lin(step):
            return 1 / self.trust_radius - 1 / np.linalg.norm(step)

        def finalize_step(shift):
            step = get_step(shift)
            if transform:
                if isinstance(eigvecs, torch.Tensor):
                    step = torch.tensor(step, device=eigvecs.device, dtype=eigvecs.dtype)
                    return (eigvecs @ step).cpu().numpy()
                else:
                    return eigvecs.dot(step)
            return step

        # Simplest case. Positive definite Hessian and predicted step is
        # already in trust radius.
        if pos_definite and newton_norm <= self.trust_radius:
            self.log("Using unshifted Newton step.")
            if isinstance(eigvecs, torch.Tensor):
                newton_step_trans = torch.tensor(
                    newton_step_trans, device=eigvecs.device, dtype=eigvecs.dtype
                )
                return (eigvecs @ newton_step_trans).cpu().numpy()
            else:
                return eigvecs.dot(newton_step_trans)

        # If the Hessian is not positive definite or if the step is too
        # long we have to determine the shift parameter lambda.
        rs_kwargs = {
            "f": lambda shift: on_trust_radius_lin(get_step(shift)),
            "xtol": 1e-3,
            # Would otherwise be chosen automatically, but we set it
            # here explicitly for verbosity.
            "method": "brentq",
        }

        def root_search(bracket):
            rs_kwargs.update(
                {
                    "bracket": bracket,
                    "x0": bracket[0] + 1e-3,
                }
            )
            res = root_scalar(**rs_kwargs)
            return res

        BRACKET_END = 1e10
        if not hard_case:
            bracket_start = 0.0 if pos_definite else -min_eigval + 1e-2
            bracket = (bracket_start, BRACKET_END)
            try:
                res = root_search(bracket)
                assert res.converged
                return finalize_step(res.root)
            # ValueError may be raised when the function values for the
            # initial bracket have the same sign. If so, we continue with
            # treating it as a hard case.
            except ValueError:
                pass

        # Now we would try the bracket (-b2, -b1). The resulting step should have
        # a suitable length, but the (shifted) Hessian would have an incorrect
        # eigenvalue spectrum (not positive definite). To solve this we use a
        # different formula to calculate the step.
        mask = np.ones_like(gradient_trans)
        mask[min_ind] = 0
        mask = mask.astype(bool)
        without_min = gradient_trans[mask] / (eigvals[mask] - min_eigval)
        tau_sq = self.trust_radius**2 - (without_min**2).sum()
        if tau_sq >= 0.0:
            tau = sqrt(tau_sq)
            step_trans = [tau] + (-without_min).tolist()
        else:
            # Hard case. Search in open interval (endpoints not included)
            # (-min_eigval, inf).
            bracket = (-min_eigval + 1e-6, BRACKET_END)
            try:
                res = root_search(bracket)
                if res.converged:
                    return finalize_step(res.root)
            except ValueError:
                pass
            # Fallback: clamp tau to 0 so the step excludes the
            # minimum-eigenvalue component but remains valid.
            self.log("Hard case fallback: tau clamped to 0.")
            tau = 0.0
            step_trans = [tau] + (-without_min).tolist()

        if not transform:
            return step_trans

        if isinstance(eigvecs, torch.Tensor):
            step_trans = torch.tensor(step_trans, device=eigvecs.device, dtype=eigvecs.dtype)
            return (eigvecs @ step_trans).cpu().numpy()
        else:
            return eigvecs.dot(step_trans)

    @staticmethod
    def quadratic_model(gradient, hessian, step):
        if isinstance(gradient, torch.Tensor):
            step = torch.tensor(step, device=gradient.device, dtype=gradient.dtype)
            return (step @ gradient + 0.5 * step @ hessian @ step).cpu().numpy()
        else:
            step = np.asarray(step).ravel()
            return step.dot(gradient) + 0.5 * step.dot(hessian).dot(step)

    @staticmethod
    def rfo_model(gradient, hessian, step):
        return HessianOptimizer.quadratic_model(gradient, hessian, step) / (
            1 + step.dot(step)
        )

    def get_step_func(self, eigvals, gradient, grad_rms_thresh=1e-2):
        positive_definite = (eigvals < 0).sum() == 0
        gradient_small = rms(gradient) < grad_rms_thresh

        if self.adapt_step_func and gradient_small and positive_definite:
            return self.get_newton_step_on_trust, self.quadratic_model
        # RFO fallback
        else:
            return self.get_rs_step, self.rfo_model



import logging

from pysisyphus.dynamics.colvars import get_colvar
from pysisyphus.dynamics.Gaussian import Gaussian
from pysisyphus.dynamics.helpers import get_mb_velocities_for_geom
from pysisyphus.dynamics.mdp import mdp
from pysisyphus.dynamics.rattle import rattle_closure
from pysisyphus.dynamics.driver import md
from pysisyphus.dynamics.wigner import get_wigner_sampler


logger = logging.getLogger("dynamics")
logger.setLevel(logging.DEBUG)
# delay = True prevents creation of empty logfiles
handler = logging.FileHandler("dynamics.log", mode="w", delay=True)
# fmt_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
fmt_str = "%(asctime)s - %(message)s"
formatter = logging.Formatter(fmt_str)
handler.setFormatter(formatter)
logger.addHandler(handler)



from collections import namedtuple
import logging
from math import sqrt
from pprint import pprint

import numpy as np
import sympy as sym

from pysisyphus.helpers_pure import log

import torch

logger = logging.getLogger("optimizer")


def gen_solutions():
    """
    Given two energies (e0, e1) and corresponding gradients (g0, g1) we
    can (try to) fit a quartic polynomial
        f(x) = a0 + a1*x + a2*x**2 + a3*x**3 + a4*x**4
    s.t. the constraint f''(x) >= 0, with the equality being fullfilled
    at only one point.
    There are five unknowns (a0 - a4) to be determined. Four equations can
    be derived from f(x) and its first derivative
        f'(x) = a1 + 2*a2*x + 3*a3*x**2 + 4*a4*x**3 .
    With (e0, g0) being given at x=0 and (e1, g1) being given at x=1 we can
    setup the following equations:
        f (0) = a0                          (1)
        f'(0) = a1                          (2)
    using e0 and g0 at x=0, and
        f (1) = a0 + a1 + a2 + a3 + a4      (3)
        f'(1) = a1 + 2*a2 + 3*a3 + 4*a4 .   (4)
    The missing last equation can be derived from the constraint. The second
    derivative of f(x) is
        f''(x) = 2*a2 + 6*a3*x + 12*a4*x**2
    and shall be positive except at one point where it is allowed to be 0, that
    its two roots (f''(x) = 0) must be degenerate. This is fullfilled when the
    discriminant D of the quadratic polynomial a*x**2 + b*x + c is zero.
        D = b**2 – 4*a*c = 0
    With
        a = 12*a4
        b =  6*a3
        c =  2*a2
    we get
        0 = (6*a3)**2 - 4*12*a4*2*a2
        0 = 36*a3**2 - 96*a4*a2
        0 = 3*a3**2 - 8*a4*a2               (5)
        or
        a4 = 3/8 * a3**2 / a2
    Using (1) - (5) we can solve the set of equations for a0 - a4.
    """

    e0, e1, g0, g1, a0, a1, a2, a3 = sym.symbols("e0 e1 g0 g1 a:4")

    a4 = sym.Rational(3, 8) * a3 ** 2 / a2
    s0, s1 = sym.solve(
        (
            e0 - a0,
            g0 - a1,
            e1 - a0 - a1 - a2 - a3 - a4,
            g1 - a1 - 2 * a2 - 3 * a3 - 4 * a4,
            3 * a3 ** 2 - 8 * a2 * a4,
        ),
        (a0, a1, a2, a3),
    )
    print("Solution 0")
    print("\t", s0)
    print()
    print("Solution 1")
    print("\t", s1)
    print()
    # There will be two solutions (s0, s1), both containing a big sqrt(...) term
    # that can be computed once and reused.
    s0_cse = sym.cse(s0)
    s1_cse = sym.cse(s1)
    print("Solution 0 after CSE")
    pprint(s0_cse)
    print("Solution 1 after CSE")
    pprint(s1_cse)
    print()
    # The terms in the sqrt-term correspond to binomial expansions and can be further
    # simplified.
    ref_term = (
        -12 * e0 ** 2
        + 24 * e0 * e1
        - 12 * e0 * g0
        - 12 * e0 * g1
        - 12 * e1 ** 2
        + 12 * e1 * g0
        + 12 * e1 * g1
        - 2 * g0 ** 2
        - 8 * g0 * g1
        - 2 * g1 ** 2
    )
    sqrt_term = -2 * (
        6 * (e0 - e1) ** 2 + 6 * (e0 - e1) * (g0 + g1) + (g0 + g1) ** 2 + 2 * g0 * g1
    )
    assert sym.simplify(sym.expand(sqrt_term) - ref_term) == 0


def get_minimum(poly):
    roots = np.roots(np.polyder(poly))
    real_roots = np.real(roots[np.isreal(roots)])
    vals = poly(real_roots)
    min_ind = vals.argmin()
    min_root = real_roots[vals.argmin()]
    min_val = vals[min_ind]
    return min_root, min_val


def get_maximum(poly):
    roots = np.roots(np.polyder(poly))
    real_roots = np.real(roots[np.isreal(roots)])
    vals = poly(real_roots)
    max_ind = vals.argmax()
    max_root = real_roots[max_ind]
    max_val = vals[max_ind]
    return max_root, max_val


FitResult = namedtuple("FitResult", "x y polys")


def quintic_fit(e0, e1, g0, g1, H0, H1):
    a = -H0 / 2 + H1 / 2 - 6 * e0 + 6 * e1 - 3 * g0 - 3 * g1
    b = 3 * H0 / 2 - H1 + 15 * e0 - 15 * e1 + 8 * g0 + 7 * g1
    c = -3 * H0 / 2 + H1 / 2 - 10 * e0 + 10 * e1 - 6 * g0 - 4 * g1
    d = H0 / 2
    e = g0
    f = e0

    poly = np.poly1d((a, b, c, d, e, f))
    try:
        mr, mv = get_minimum(poly)
    except ValueError:
        return None

    fit_result = FitResult(mr, mv, (poly,))
    return fit_result


def quartic_fit(e0, e1, g0, g1, maximize=False):
    """See gen_solutions() for derivation."""
    a0 = e0
    a1 = g0
    try:
        sqrt_term = sqrt(
            -2
            * (
                6 * (e0 - e1) ** 2
                + 6 * (e0 - e1) * (g0 + g1)
                + (g0 + g1) ** 2
                + 2 * g0 * g1
            )
        )
    except ValueError:
        # In these cases there is no intermediate minimum between 0 and 1 and the term
        # under the square root becomes negative.
        return None

    a2_pre = -3 * (e0 - e1) - 5 * g0 / 2 - g1 / 2
    a3_pre = 2 * e0 - 2 * e1 + 2 * g0

    def get_poly(a3, a2, a1, a0):
        a4 = 3 / 8 * a3 ** 2 / a2
        return np.poly1d((a4, a3, a2, a1, a0))

    a2 = a2_pre - sqrt_term / 2
    a3 = a3_pre + sqrt_term
    poly0 = get_poly(a3, a2, a1, a0)

    a2 = a2_pre + sqrt_term / 2
    a3 = a3_pre - sqrt_term
    poly1 = get_poly(a3, a2, a1, a0)

    get_func = get_maximum if maximize else get_minimum
    mr0, mv0 = get_func(poly0)
    mr1, mv1 = get_func(poly1)

    if maximize:
        mr, mv = (mr0, mv0) if mv0 > mv1 else (mr1, mv1)
    else:
        mr, mv = (mr0, mv0) if mv0 < mv1 else (mr1, mv1)

    # Shorter sympy implementation. Probably slower? But shouldn't matter...
    # ... of course it does ;)
    # a0, a1, a2, a3 = sym.symbols("a:4")
    # a4 = sym.Rational(3, 8) * a3**2 / a2
    # s0, s1 = sym.solve((e0-a0,
    # g0-a1,
    # e1-a0-a1-a2-a3-a4,
    # g1-a1-2*a2-3*a3-4*a4,
    # 3*a3**2 - 8*a2*a4),
    # (a0, a1, a2, a3)
    # )
    # N = lambda exprs: [sym.N(expr) for expr in exprs]
    # sym_poly0 = get_poly(*N(s0[::-1]))
    # sym_poly1 = get_poly(*N(s1[::-1]))

    fit_result = FitResult(mr, mv, (poly0, poly1))
    return fit_result


def cubic_fit(e0, e1, g0, g1):
    # # Shorter sympy implementation. Probably slower? But shouldn't matter...
    # # Ok it is really slow ... and it's gone.
    # a0, a1, a2, a3 = sym.symbols("a:4")
    # s = sym.solve((e0-a0,
    # g0-a1,
    # e1-a0-a1-a2-a3,
    # g1-a1-2*a2-3*a3),
    # (a0, a1, a2, a3),
    # )
    # coeffs = [float(sym.N(expr)) for expr in (s[a3], s[a2], s[a1], s[a0])]
    d = e0
    c = g0
    b = -(g1 + 2 * g0 + 3 * e0 - 3 * e1)
    a = 2 * (e0 - e1) + g0 + g1
    # np.testing.assert_allclose([a, b, c, d], coeffs, atol=1e-10)
    poly = np.poly1d((a, b, c, d))
    try:
        mr, mv = get_minimum(poly)
    except ValueError:
        return None

    fit_result = FitResult(mr, mv, (poly,))
    return fit_result


def poly_line_search(
    cur_energy,
    prev_energy,
    cur_grad,
    prev_grad,
    prev_step,
    cubic_max_x=2.0,
    quartic_max_x=4.0,
    logger=None,
):
    """Generate directional gradients by projecting them on the previous step."""
    return_torch = False
    if isinstance(cur_grad, torch.Tensor):
        return_torch = True
        device, dtype = cur_grad.device, cur_grad.dtype
        cur_grad = cur_grad.cpu().numpy()
    prev_grad_proj = prev_step @ prev_grad
    cur_grad_proj = prev_step @ cur_grad
    cubic_result = cubic_fit(prev_energy, cur_energy, prev_grad_proj, cur_grad_proj)
    quartic_result = quartic_fit(prev_energy, cur_energy, prev_grad_proj, cur_grad_proj)
    accept = {
        "cubic": lambda x: (x > 0.0) and (x < cubic_max_x),
        "quartic": lambda x: (x > 0.0) and (x <= quartic_max_x),
    }

    fit_result = None
    if quartic_result and accept["quartic"](quartic_result.x):
        fit_result = quartic_result
        deg = "quartic"
    elif cubic_result and accept["cubic"](cubic_result.x):
        fit_result = cubic_result
        deg = "cubic"

    fit_energy = None
    fit_grad = None
    fit_step = None
    if fit_result and fit_result.y < prev_energy:
        x = fit_result.x
        fit_energy = fit_result.y
        log(logger, f"Did {deg} interpolation with x={x:.6f}.")

        # Interpolate coordinates and gradient. 'fit_step' applied to the current
        # coordinates yields interpolated coordinates.
        #
        # x == 0 would take us to the previous coordinates:
        #  (1-0) * -prev_step = -prev_step (we revert the last step)
        # x == 1 would preserve the current coordinates:
        #  (1-1) * -prev_step = 0 (we stay at the current coordinates)
        # x > 1 extrapolate along previous step direction:
        #  with x=2, (1-2) * -prev_step = -1*-prev_step = prev_step
        fit_step = (1 - x) * -prev_step
        fit_grad = (1 - x) * prev_grad + x * cur_grad
        if return_torch:
            fit_grad = torch.tensor(fit_grad, dtype=dtype, device=device)
            fit_step = torch.tensor(fit_step, dtype=dtype, device=device)
    return fit_energy, fit_grad, fit_step



import numpy as np
from sympy import symbols, exp

# [1]   10.1142/9789812839664_0016
#       G. Henkelman, G. Jóhannesson, and H. Jónsson
#       Theoretical Methods in Condensed Phase Chemistry
#       (Springer, Netherlands, 2002), pp. 287
# [2]   https://aip.scitation.org/doi/pdf/10.1063/1.480097?class=pdf
#       I'm very sure that Fig. 6 and Fig. 5 are not the potentials that are
#       described in the paper. There are two errors in the paper. A_1 is
#       actually -1.5 and not 1.5 and s_x_2 is 0.5 and not 5.0. You also
#       have to divide by (2*s_x_i**2) and not only by (2*s_x_i) as given
#       in Eq. (22).

class LEPSExpr:


    def __init__(self):
        """Generates sympy expression for various LEPS potentials."""
        # V_harmonic uses b = 0.80
        self.abc = (0.05, 0.30, 0.05)
        self.ds = (4.746, 4.746, 3.445)
        self.r0 = 0.742
        self.alpha = 1.942

        # Used for LEPS + harmonic oscillator
        self.rac = 3.742
        self.kc = 0.2025
        self.c_ = 1.154

        self.choices = {
            "leps": self.get_leps,
            "harmonic": self.get_harmonic,
            "tot": self.get_tot,
            "dimer": self.get_dimer,
        }

    def get_leps(self):
        V_expr = self.V_LEPS()
        xlim = (0, 6)
        ylim = (0, 12)
        levels = np.linspace(-5, 5, 250)
        return V_expr, xlim, ylim, levels

    def get_harmonic(self):
        V_expr = self.V_harmonic()
        levels = np.linspace(-20, 20, 250)
        xlim = (0.4, 3.2)
        ylim = (-2, 2)
        return V_expr, xlim, ylim, levels

    def get_tot(self):
        V_expr = self.V_tot()
        levels = np.linspace(-5, 2, 75)
        xlim = (0.4, 3.2)
        ylim = (-2, 2)
        return V_expr, xlim, ylim, levels

    def get_dimer(self):
        V_expr = self.V_dimer()
        levels = None
        xlim = (0.25, 3.25)
        ylim = (-3.5, 3.5)
        return V_expr, xlim, ylim, levels

    def get_expr(self, pot_type="leps"):
        assert pot_type in self.choices
        return self.choices[pot_type]()


    def Q(self, d, alpha, r0, r):
        """Coulomb interactions."""
        return d/2* (3/2*exp(-2*alpha*(r-r0)) - exp(-alpha*(r-r0)))

    def J(self, d, alpha, r0, r):
        """Quantum mechanical exchange interaction."""
        return d/4 * (exp(-2*alpha*(r-r0))-6*exp(-alpha*(r-r0)))

    def G(self, a, b):
        "Gaussian function."""
        return exp(-0.5*((a/0.1)**2 +(b/0.35)**2))

    def Gdimer(self, x, y, A, x0, y0, sx, sy):
        return A * exp(-(x-x0)**2/(2*sx**2)) * exp(-(y-y0)**2/(2*sy**2))

    def V_LEPS(self, x=None, y=None, abc=None):
        """Equation (A.1) in [1].
        Mimics reaction involving three atoms confined to motion along
        a line."""
        if abc is None:
            abc = self.abc
        a, b, c = abc
        if x is None:
            x = symbols("x")
        if y is None:
            y = symbols("y")
        rs = (x, y, self.rac)
        
        Qab, Qbc, Qac = [self.Q(d, self.alpha, self.r0, r)
                         for d, r in zip(self.ds, rs)]
        Jab, Jbc, Jac = [self.J(d, self.alpha, self.r0, r)
                         for d, r in zip(self.ds, rs)]
        
        first_term = Qab/(1+a) + Qbc/(1+b) + Qac/(1+c)
        second_term = Jab**2/(1+a)**2 + Jbc**2/(1+b)**2 + Jac**2/(1+c)**2
        third_term = (-Jab*Jbc / ((1+a)*(1+b))
                      -Jbc*Jac / ((1+b)*(1+c))
                      -Jab*Jac / ((1+a)*(1+c))
        )
        return first_term - (second_term + third_term)**0.5

    def V_harmonic(self):
        """Equation (A.2) in [1].
        A and C are fixed, only B can move. A condensed phase environment
        is represented by adding a harmonic oscillator degree of freedom."""
        x, y = symbols("x y")
        abc = (0.05, 0.80, 0.05)
        return (self.V_LEPS(x, self.rac-x, abc)
                + 2*self.kc*(x-(self.rac/2 - y/self.c_))**2
        )

    def V_tot(self):
        """Equation (A.3) in [1].
        Additional saddle point."""
        x, y = symbols("x y")
        return self.V_harmonic() + 1.5 * self.G(x-2.02083, y+0.272881)

    def V_dimer(self):
        """III. Results Section A in [3]. Two additional saddle points
        from two added gaussians."""
        x, y = symbols("x y")
        #           Ai , x0i    ,  y0i     , sxi, syi
        params = ((-1.5, 2.02083, -0.172881, 0.1, 0.35),
                  ( 6.0, 0.8    ,  2.0     , 0.5, 0.7 ))
        G0, G1 = [self.Gdimer(x, y, *parms) for parms in params]
        return self.V_harmonic() + G0 + G1

    def __str__(self):
        return "LEPSExpr"



import re

import jinja2
import numpy as np

from pysisyphus.constants import ANG2BOHR
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers_pure import file_or_str


@file_or_str(".crd")
def parse_crd(crd):
    lines = [l.strip() for l in crd.strip().split("\n") if not l.startswith("*")]
    expect_line = lines.pop(0)
    expect_num = int(expect_line.split()[0])
    coords3d = np.zeros((expect_num, 3), dtype=float)
    atoms = list()
    re_ = re.compile(r"\d+")
    for i, line in enumerate(lines):
        _, _, res, atom, x, y, z, *_ = line.split()
        coords3d[i] = (x, y, z)
        atom = re_.sub("", atom)  # Delte number
        atoms.append(atom)
    assert len(atoms) == expect_num
    coords3d *= ANG2BOHR
    return atoms, coords3d.flatten()


def geom_from_crd(fn, **kwargs):
    atoms, coords = parse_crd(fn)
    return Geometry(atoms, coords, **kwargs)


# I10   I10 2X A8 2X A8       3F20.10     2X A8 2X A8 F20.10
CRD_TPL = """* Created by pysisyphus
*
{{ i10(atomnum) }}  EXT
{% for atom, x, y, z in atoms_xyz -%}
 {{ i10(loop.index) }}{{ i10(resno) }}  {{ a8(res) }}  {{ a8(atom) }}{{ f20(x) }}{{ f20(y) }}{{ f20(z) }}  {{ a8(segid) }}  {{ a8(resid) }}{{ f20(0.0) }}
{% endfor %}
"""


def i10(int_):
    return f"{int_:10d}"


def f20(float_):
    return f"{float_:>20.10f}"


def a8(str_):
    return f"{str(str_):<8s}"


def atoms_coords_to_crd_str(
    atoms,
    coords,
    resno=1,
    res="UNL1",
    segid=None,
    resid=1,
    ref_atoms=None,
    del_atoms=None,
):
    if segid is None:
        segid = res
    if del_atoms is None:
        del_atoms = []

    env = jinja2.Environment(loader=jinja2.BaseLoader)
    env.globals.update(i10=i10, f20=f20, a8=a8)
    tpl = env.from_string(CRD_TPL)

    if ref_atoms is not None:
        atoms_for_names = ref_atoms
    else:
        atoms_for_names = atoms
    counter = dict()
    atom_names = list()
    for atom in atoms_for_names:
        counter.setdefault(atom, 1)
        atom_names.append(f"{atom}{counter[atom]}")
        counter[atom] += 1
    atom_names = [an for i, an in enumerate(atom_names) if i not in del_atoms]

    coords3d = coords.reshape(-1, 3) / ANG2BOHR
    atoms_xyz = list(zip(atom_names, *coords3d.T))
    rendered = tpl.render(
        atomnum=len(atoms_xyz),
        atoms_xyz=atoms_xyz,
        resno=resno,
        res=res,
        resid=resid,
        segid=segid,
    )
    return rendered


def geom_to_crd_str(geom, **kwargs):
    return atoms_coords_to_crd_str(geom.atoms, geom.cart_coords, **kwargs)



import numpy as np
from scipy.interpolate import splprep, splev

from pysisyphus.constants import AU2KCALPERMOL
from pysisyphus.intcoords.exceptions import (
    DifferentCoordLengthsException,
    DifferentPrimitivesException,
    RebuiltInternalsException,
)
from pysisyphus.cos.ChainOfStates import ChainOfStates
from pysisyphus.cos.GrowingChainOfStates import GrowingChainOfStates


# [1] https://aip.scitation.org/doi/abs/10.1063/1.1691018
#     Peters, 2004
# [2] https://aip.scitation.org/doi/abs/10.1063/1.4804162
#     Zimmerman, 2013


class GrowingString(GrowingChainOfStates):
    def __init__(
        self,
        images,
        calc_getter,
        perp_thresh=0.05,
        param="equi",
        reparam_every=2,
        reparam_every_full=3,
        reparam_tol=None,
        reparam_check="rms",
        max_micro_cycles=5,
        reset_dlc=True,
        climb=False,
        **kwargs,
    ):
        assert len(images) >= 2, "Need at least 2 images for GrowingString."
        if len(images) > 2:
            images = [images[0], images[-1]]
            print("More than 2 images given. Will only use first and last image!")
        if climb:
            climb = "one"
        super().__init__(images, calc_getter, climb=climb, **kwargs)

        self.perp_thresh = perp_thresh
        self.param = param
        self.reparam_every = int(reparam_every)
        self.reparam_every_full = int(reparam_every_full)
        assert (
            self.reparam_every >= 1 and self.reparam_every_full >= 1
        ), "reparam_every and reparam_every_full must be positive integers!"
        if reparam_tol is not None:
            self.reparam_tol = float(reparam_tol)
            assert self.reparam_tol > 0
        else:
            self.reparam_tol = 1 / (self.max_nodes + 2) / 2
        self.log(f"Using reparametrization tolerance of {self.reparam_tol:.4f}")
        self.reparam_check = reparam_check
        assert self.reparam_check in ("norm", "rms")
        self.max_micro_cycles = int(max_micro_cycles)
        self.reset_dlc = bool(reset_dlc)

        left_img, right_img = self.images

        self.left_string = [
            left_img,
        ]
        self.right_string = [
            right_img,
        ]

        # The desired spacing of the nodes in the final string on the
        # normalized arclength.
        self.sk = 1 / (self.max_nodes + 1)

        self.reparam_in = reparam_every
        self._tangents = None
        self.tangent_list = list()
        self.perp_forces_list = list()
        self.coords_list = list()

        left_frontier = self.get_new_image(self.lf_ind)
        self.left_string.append(left_frontier)
        right_frontier = self.get_new_image(self.rf_ind)
        self.right_string.append(right_frontier)
        self.new_image_inds = list()

    def get_cur_param_density(self, kind=None):
        diffs = [
            image - self.images[max(i - 1, 0)] for i, image in enumerate(self.images)
        ]

        norms = np.linalg.norm(diffs, axis=1)
        param_density = np.cumsum(norms)
        self.log(f"Current string length={param_density[-1]:.6f}")

        # Energy weighted parametrization density
        if kind == "energy":
            prev_energies = np.array(self.all_energies[-1])

            if len(prev_energies) != len(self.images):
                return None

            mean_energies = (prev_energies[1:] + prev_energies[:-1]) / 2
            weights = mean_energies - prev_energies.min()
            # This damps everything a bit.
            weights = np.sqrt(weights)
            param_density = [
                0,
            ]
            for weight, diff in zip(weights, norms[1:]):
                assert weight > 0.0
                param_density.append(param_density[-1] + weight * diff)

        param_density = np.array(param_density)
        param_density /= param_density[-1]

        return param_density

    def reset_geometries(self, ref_geometry):
        ref_typed_prims = ref_geometry.internal.typed_prims
        self.log(
            f"Resetting image primitives. Got {len(ref_typed_prims)} typed primitives."
        )
        # Do multiple cycles as it may happen, that not all coordinates are valid
        # at every node.
        for i in range(3):
            self.log(f"\tMicro cycle {i:d}")
            intersect = set(self.images[0].internal.typed_prims)
            for j, image in enumerate(self.images):
                image.reset_coords(ref_typed_prims)
                new_typed_prims = set(image.internal.typed_prims)
                self.log(
                    f"\tImage {j:02d} now has {len(new_typed_prims)} typed primitives."
                )
                intersect = intersect & new_typed_prims

            if intersect == set(ref_typed_prims):
                ref_geometry.reset_coords(intersect)
                break
            ref_typed_prims = list(intersect)
        else:
            raise Exception("Too many reset cycles!")

    def set_coords(self, image, coords):
        try:
            image.coords = coords
        except RebuiltInternalsException:
            print("Rebuilt internal coordinates!")
            self.reset_geometries(image)

    def get_new_image(self, ref_index):
        """Get new image by taking a step from self.images[ref_index] towards
        the center of the string."""
        new_img = self.images[ref_index].copy(
            coord_kwargs={
                "check_bends": True,
            }
        )

        if ref_index <= self.lf_ind:
            tangent_ind = ref_index + 1
            insert_ind = tangent_ind
        else:
            tangent_ind = ref_index - 1
            insert_ind = ref_index
        tangent_img = self.images[tangent_ind]

        # (new_img - tangent_img) points from tangent_img towards new_img.
        # As we want to derive a new image from new_img, we have to step
        # against this vector, so we have to multiply by -1.
        # Why don't we just use (tangent_img - new_img) to get the right
        # direction? In DLC the resulting distance would then be given in
        # the active set U of tangent_img, but we need it in the active set U
        # of new_img.
        # Formulated the other way around the same expression can be used for
        # all coord types.
        try:
            distance = -(new_img - tangent_img)
        except (DifferentCoordLengthsException, DifferentPrimitivesException):
            self.reset_geometries(new_img)
            distance = -(new_img - tangent_img)

        # The desired step(_length) for the new image be can be easily determined
        # from a simple rule of proportion by relating the actual distance between
        # two images to their parametrization density difference on the normalized
        # arclength and the desired spacing given by self.sk.
        #
        # Δparam_density / distance = self.sk / step
        # step = self.sk / Δparam_density * distance
        cpd = self.get_cur_param_density()
        # As we always want to step in the direction of 'distance' we just take
        # the absolute value of the difference, as we are not interested in the
        # sign.
        param_dens_diff = abs(cpd[ref_index] - cpd[tangent_ind])
        step_length = self.sk / param_dens_diff
        step = step_length * distance

        new_coords = new_img.coords + step
        self.set_coords(new_img, new_coords)

        new_img.set_calculator(self.calc_getter())
        ref_calc = self.images[ref_index].calculator
        try:
            chkfiles = ref_calc.get_chkfiles()
            new_img.calculator.set_chkfiles(chkfiles)
            self.log(
                "Set checkfiles from calculator of node "
                f"{ref_index:02d} on calculator of new node."
            )
        except AttributeError:
            self.log("Calculator doesn't support 'get/set_chkfiles()'")
        self.images.insert(insert_ind, new_img)
        self.log(f"Created new image; inserted it before index {insert_ind}.")
        return new_img

    @property
    def left_size(self):
        return len(self.left_string)

    @property
    def right_size(self):
        return len(self.right_string)

    @property
    def string_size(self):
        return self.left_size + self.right_size

    @property
    def fully_grown(self):
        """Returns wether the string is fully grown. Don't count the first
        and last node."""
        return not ((self.string_size - 2) < self.max_nodes)

    @property
    def nodes_missing(self):
        """Returns the number of nodes to be grown."""
        return (self.max_nodes + 2) - self.string_size

    @property
    def lf_ind(self):
        """Index of the left frontier node in self.images."""
        return len(self.left_string) - 1

    @property
    def rf_ind(self):
        """Index of the right frontier node in self.images."""
        return self.lf_ind + 1

    @property
    def full_string_image_inds(self):
        left_inds = np.arange(self.left_size)
        right_inds = np.arange(self.max_nodes + 2)[-self.right_size :]
        image_inds = np.concatenate((left_inds, right_inds))
        return image_inds

    @property
    def image_inds(self):
        return self.full_string_image_inds

    def spline(self, tangents=False):
        if (not tangents) and (self.param == "energy") and self.fully_grown:
            u = self.get_cur_param_density(kind="energy")
        else:
            u = self.get_cur_param_density()
        reshaped = self.coords.reshape(-1, self.coords_length)
        # To use splprep we have to transpose the coords.
        transp_coords = reshaped.transpose()
        # Spline in batches as scipy can't handle > 11 rows at once
        tcks, us = zip(
            *[
                splprep(transp_coords[i : i + 9], s=0, k=3, u=u)
                for i in range(0, len(transp_coords), 9)
            ]
        )
        return tcks, us

    def reparam_cart(self, desired_param_density):
        tcks, us = self.spline()
        # Reparametrize mesh
        new_points = np.vstack([splev(desired_param_density, tck) for tck in tcks])
        # Flatten along first dimension.
        new_points = new_points.reshape(-1, len(self.images)).T
        # With a climbing image we ignore the just splined coordinates for the CI
        # and restore its original coordinates.
        for index in self.get_climbing_indices():
            new_points[index] = self.images[index].coords
            self.log(f"Skipped reparametrization of climbing image with index {index}")
        self.coords = new_points.flatten()
        # In contrast to self.reparam_dlc() we don't check if the reparametrization
        # succeeded because it can't fail ;)

    def reparam_dlc(self, desired_param_density, thresh=1e-3):
        climbing_indices = self.get_climbing_indices()
        # Reparametrization will take place along the tangent between two
        # images. The index of the tangent image depends on wether the image
        # is above or below the desired param_density on the normalized arc.
        #
        # The reparametrization is done in micro cycles, until it is converged.
        cur_param_density = self.get_cur_param_density()
        self.log(f"Density before reparametrization: {cur_param_density}")
        for i, reparam_image in enumerate(self.images[1:-1], 1):
            if i in climbing_indices:
                self.log(f"Skipped reparametrization of climbing image with index {i}")
                continue
            self.log(f"Reparametrizing node {i}")
            for j in range(self.max_micro_cycles):
                diff = (desired_param_density - cur_param_density)[i]
                self.log(f"\t{j}: Δ={diff: .6f}")
                # Do at least one pass
                if (j > 0) and (abs(diff) < thresh):
                    break
                # Negative sign: image is too far right and has to be shifted left.
                # Positive sign: image is too far left and has to be shifted right.
                sign = int(np.sign(diff))
                # Index of the tangent image. reparam_image will be shifted along
                # this direction to achieve the desired parametirzation density.
                tangent_ind = i + sign
                tangent_image = self.images[tangent_ind]
                rl = "right" if sign > 0 else "left"
                self.log(f"\t... shifting {rl} towards image {tangent_ind}")
                distance = -(reparam_image - tangent_image)

                param_dens_diff = abs(
                    cur_param_density[tangent_ind] - cur_param_density[i]
                )
                step_length = abs(diff) / param_dens_diff
                step = step_length * distance
                reparam_coords = reparam_image.coords + step
                self.set_coords(reparam_image, reparam_coords)
                cur_param_density = self.get_cur_param_density()
            else:
                self.log(
                    f"Reparametrization of node {i} did not converge after "
                    f"{self.max_micro_cycles} cycles. Breaking!"
                )
                break

        cpd_str = np.array2string(cur_param_density, precision=4)
        self.log(f"Param density after reparametrization: {cpd_str}")

        # This check is disabled at it is not really applicable. While we reparametrize
        # the images the string size may vary wildly, at least in the beginning. Lets
        # say after reparametrization the distance vector between image 0 and 1 is of
        # magnitude 1 and the overall string length is 10. Then image 1 is at 0.1 w.r.t.
        # the parametrization density. If we reparametrize the remaining images the over-
        # all string size may be 8, and now image 1 suddenly sits at 1/8 = 0.125, which
        # may be already above the allowed threshold.
        # Over time the string size will equilibrate and the desired parametrization
        # density will actually be realized.
        # try:
        # # Dont check climbing images
        # np.testing.assert_allclose(
        # np.delete(cur_param_density, climbing_indices),
        # np.delete(desired_param_density, climbing_indices),
        # atol=self.reparam_tol
        # )
        # except AssertionError as err:
        # trj_str = self.as_xyz()
        # fn = "failed_reparametrization_trj.xyz"
        # with open(fn, "w") as handle:
        # handle.write(trj_str)
        # print(f"Wrote coordinates of failed reparametrization to '{fn}'")
        # raise err

        # Regenerate active set after reparametrization
        if self.reset_dlc and not self.fully_grown:
            [image.internal.set_active_set() for image in self.moving_images]
            self.log(f"Created new DLCs for {len(self.images)} string images.")
        elif self.reset_dlc:
            self.log("Skipping creation of new DLCs, as string is already fully grown.")

    def get_tangent(self, i):
        # Simple tangent, pointing at each other, for the frontier images.
        if not self.fully_grown and i in (self.lf_ind, self.rf_ind):
            next_ind = i + 1 if (i <= self.lf_ind) else i - 1
            tangent = self.images[next_ind] - self.images[i]
            tangent /= np.linalg.norm(tangent)
        else:
            tangent = super().get_tangent(i, kind="upwinding")

        return tangent

    @ChainOfStates.forces.getter
    def forces(self):
        if self._forces is None:
            self.calculate_forces()
        indices = range(len(self.images))
        # In constrast to NEB calculations we only use the perpendicular component
        # of the force, without any spring forces. A desired image distribution is
        # achieved via periodic reparametrization.
        perp_forces = np.array([self.get_perpendicular_forces(i) for i in indices])
        self.perp_forces_list.append(perp_forces.copy().flatten())
        # Add climbing forces
        total_forces = self.set_climbing_forces(perp_forces)
        self._forces = total_forces.flatten()
        return self._forces

    def reparametrize(self):
        reparametrized = False
        # If this counter reaches 0 reparametrization will occur.
        self.reparam_in -= 1

        self.new_image_inds = list()
        # Check if new images can be added for incomplete strings.
        if not self.fully_grown:
            perp_forces = self.perp_forces_list[-1].reshape(len(self.images), -1)
            # Calculate norm and rms of the perpendicular force for every
            # node/image on the string.
            to_check = {
                "norm": np.linalg.norm(perp_forces, axis=1),
                "rms": np.sqrt(np.mean(perp_forces**2, axis=1)),
            }
            self.log(
                f"Checking frontier node convergence, threshold={self.perp_thresh:.6f}"
            )
            # We can add a new node if the norm/rms of the perpendicular force is below
            # the threshold.
            def converged(i):
                cur_val = to_check[self.reparam_check][i]
                is_converged = cur_val <= self.perp_thresh
                conv_str = ", converged" if is_converged else ""
                self.log(
                    f"\tnode {i:02d}: {self.reparam_check}(perp_forces)={cur_val:.6f}"
                    f"{conv_str}"
                )
                return is_converged

            # New images are added with the same coordinates as the frontier image.
            # We force reparametrization by setting self.reparam_in to 0 to get sane
            # coordinates for the new image(s).
            if converged(self.lf_ind):
                # Insert at the end of the left string, just before the
                # right frontier node.
                new_left_frontier = self.get_new_image(self.lf_ind)
                self.new_image_inds.append(self.left_size)
                self.left_string.append(new_left_frontier)
                self.log("Added new left frontier node.")
                self.reparam_in = 0
            # If an image was just grown in the left substring the string may now
            # be fully grown, so we reavluate 'self.fully_grown' here.
            if (not self.fully_grown) and converged(self.rf_ind):
                # Insert at the end of the right string, just before the
                # current right frontier node.
                new_right_frontier = self.get_new_image(self.rf_ind)
                self.new_image_inds.append(self.left_size)
                self.right_string.append(new_right_frontier)
                self.log("Added new right frontier node.")
                self.reparam_in = 0
            self.log(f"New image indices: {self.new_image_inds}")

        self.log(
            f"Current string size is {self.left_size}+{self.right_size}="
            f"{self.string_size}. There are still {self.nodes_missing} "
            "nodes to be grown."
            if not self.fully_grown
            else "String is fully grown."
        )

        if self.reparam_in > 0:
            self.log(
                "Skipping reparametrization. Next reparametrization in "
                f"{self.reparam_in} cycles."
            )
        else:
            # Prepare image reparametrization
            desired_param_density = self.sk * self.full_string_image_inds
            pd_str = np.array2string(desired_param_density, precision=4)
            self.log(f"Desired param density: {pd_str}")

            # Reparametrize images.
            if self.coord_type == "cart":
                self.reparam_cart(desired_param_density)
            elif self.coord_type == "dlc":
                self.reparam_dlc(desired_param_density, thresh=self.reparam_tol)
            else:
                raise Exception("How did you get here?")

            self.reparam_in = (
                self.reparam_every_full if self.fully_grown else self.reparam_every
            )
            reparametrized = True
            # Writing is deactivated, as this does not respect an out_dir or
            # something similar.
            # with open("reparametrized_trj.xyz", "w") as handle:
            # handle.write(self.as_xyz())

        return reparametrized

    def get_additional_print(self):
        size_str = f"{self.left_size}+{self.right_size}"
        if self.fully_grown:
            size_str = " Full"
        size_info = f"String={size_str: >5s}"
        energies = np.array(self.all_energies[-1])
        barrier = (energies.max() - energies[0]) * AU2KCALPERMOL
        barrier_info = f"(E_hei-E_0)={barrier:6.1f} kcal/mol"
        hei_ind = energies.argmax()
        hei_norm = np.linalg.norm(self.all_true_forces[-1][hei_ind])
        hei_info = f"norm(forces_true,hei)={hei_norm:.6f} E_h/a_0"
        hei_str = f"HEI={hei_ind+1:02d}/{energies.size:02d}"

        strs = (
            size_info,
            hei_str,
            barrier_info,
            hei_info,
        )
        return "\t" + " ".join(strs)



import itertools as it

import numpy as np


def fin_diff_prim(primitive, coords3d, delta=1e-6):
    """Derivatives of a primitive internal gradient wrt its defining
    cartesian coordinates."""
    displacement_inds = [(i, j) for i, j in it.product(primitive.indices, (0, 1, 2))]

    prim_grads = list()
    for atom_ind, ax_ind in displacement_inds:
        plus = coords3d.copy()
        plus[atom_ind, ax_ind] += delta
        plus_val = primitive.calculate(plus)
        minus = coords3d.copy()
        minus[atom_ind, ax_ind] -= delta
        minus_val = primitive.calculate(minus)
        # Select the appropriate item of the primitive internal gradient
        prim_grad = (plus_val - minus_val) / (2 * delta)
        prim_grads.append(prim_grad)
    prim_grads = np.array(prim_grads)
    return prim_grads


def fin_diff_B(primitive, coords3d, delta=1e-6):
    """Derivatives of a primitive internal gradient wrt its defining
    cartesian coordinates."""
    displacement_inds = [(i, j) for i, j in it.product(primitive.indices, (0, 1, 2))]

    B_grads = list()
    for grad_atom_ind, grad_ax_ind in displacement_inds:
        # Calculate the derivative of an entry of the Wilson B-matrix.
        for atom_ind, ax_ind in displacement_inds:
            plus = coords3d.copy()
            plus[atom_ind, ax_ind] += delta
            _, plus_val = primitive.calculate(plus, gradient=True)
            plus_val = plus_val.reshape(-1, 3)
            minus = coords3d.copy()
            minus[atom_ind, ax_ind] -= delta
            _, minus_val = primitive.calculate(minus, gradient=True)
            minus_val = minus_val.reshape(-1, 3)
            # Select the appropriate item of the primitive internal gradient
            B_grad = (plus_val[grad_atom_ind, grad_ax_ind]
                      - minus_val[grad_atom_ind, grad_ax_ind]) / (2 * delta)
            B_grads.append(B_grad)
    B_grads = np.array(B_grads)
    return B_grads



# [1] http://dx.doi.org/10.1063/1.4952956
#     Lee-Ping Wang, 2016

import numpy as np

from pysisyphus.intcoords.Primitive import Primitive
from pysisyphus.linalg import eigvec_grad, rot_quaternion


def compare_to_geometric(c3d, ref_c3d, dR, dF, dqdx, dvdx, atol=1e-14):
    from geometric.rotate import get_R_der, get_F_der, get_q_der, get_expmap_der

    dR_ref = get_R_der(c3d, ref_c3d)
    np.testing.assert_allclose(dR, dR_ref)
    dF_ref = get_F_der(c3d, ref_c3d)
    np.testing.assert_allclose(dF.reshape(-1, 3, 4, 4), dF_ref)
    dq_ref = get_q_der(c3d, ref_c3d)
    np.testing.assert_allclose(dqdx.T.flatten(), dq_ref.flatten(), atol=atol)
    dvdx_ref = get_expmap_der(c3d, ref_c3d)
    np.testing.assert_allclose(dvdx.T.flatten(), dvdx_ref.flatten(), atol=atol)


class Rotation(Primitive):
    """See (II. Theory) in [1], Eq. (3) - (14)"""

    index = None

    def __init__(self, indices, *args, ref_coords3d, **kwargs):
        kwargs["cache"] = False
        kwargs["calc_kwargs"] = ("index", "ref_coords3d")
        super().__init__(indices, *args, **kwargs)

        self.ref_coords3d = ref_coords3d.reshape(-1, 3).copy()

    @staticmethod
    def _weight(atoms, coords3d, indices, f_damping):
        return 1

    @staticmethod
    def _calculate(coords3d, indices, gradient=False, index=0, ref_coords3d=None):
        w, v_, c3d, ref_c3d = rot_quaternion(coords3d[indices], ref_coords3d[indices])

        quat = v_[:, -1]
        # Eigenvector sign is ambigous. Force first item to be positive,
        # similar to geomeTRIC code.
        if quat[0] < 0.0:
            quat *= -1

        # Eq. (8) in [1].
        # v = 2 * q_i * (cos⁻¹(q_0) / sqrt(1 - q_0 ** 2)
        #
        # As q_0 approaches 1, the denominator becomes very small, and dividing
        # by this small number results in numerical instability.
        #
        # According to wolframalpha v(q_0) limit approaches 2 for q_0 = 1.
        #
        #   input: limit of (2 * arccos(x) / sqrt(1-x**2))
        #   output: lim v(x) for x -> 1 becomes 2.
        q0 = quat[0]
        if abs(q0 - 1.0) <= 1e-8:
            prefac = 2 - 2 / 3 * (q0 - 1)
            dvdq0 = -2 / 3
        else:
            arccos_q0 = np.arccos(q0)
            diff = 1 - q0 ** 2
            prefac = 2 * arccos_q0 / np.sqrt(diff)
            dvdq0 = 2 * q0 * arccos_q0 / diff ** 1.5 - 2 / diff

        # Exponential map
        v = prefac * quat[1:]

        if gradient:
            # Gradient of correlation matrix
            y1, y2, y3 = ref_c3d.T
            dR = np.zeros((*c3d.shape, 3, 3))
            dR[:, 0, 0, 0] = y1
            dR[:, 0, 0, 1] = y2
            dR[:, 0, 0, 2] = y3
            #
            dR[:, 1, 1, 0] = y1
            dR[:, 1, 1, 1] = y2
            dR[:, 1, 1, 2] = y3
            #
            dR[:, 2, 2, 0] = y1
            dR[:, 2, 2, 1] = y2
            dR[:, 2, 2, 2] = y3
            dR11, dR12, dR13, dR21, dR22, dR23, dR31, dR32, dR33 = dR.reshape(-1, 9).T

            # Gradient of F matrix. Construct full matrix, as we have to do a dot
            # product later on.
            dF = np.zeros((ref_c3d.size, 4, 4))
            dF[:, 0, 0] = dR11 + dR22 + dR33
            dF[:, 0, 1] = dR23 - dR32
            dF[:, 0, 2] = dR31 - dR13
            dF[:, 0, 3] = dR12 - dR21
            #
            dF[:, 1, 0] = dF[:, 0, 1]
            dF[:, 1, 1] = dR11 - dR22 - dR33
            dF[:, 1, 2] = dR12 + dR21
            dF[:, 1, 3] = dR13 + dR31
            #
            dF[:, 2, 0] = dF[:, 0, 2]
            dF[:, 2, 1] = dF[:, 1, 2]
            dF[:, 2, 2] = -dR11 + dR22 - dR33
            dF[:, 2, 3] = dR23 + dR32
            #
            dF[:, 3, 0] = dF[:, 0, 3]
            dF[:, 3, 1] = dF[:, 1, 3]
            dF[:, 3, 2] = dF[:, 2, 3]
            dF[:, 3, 3] = -dR11 - dR22 + dR33

            # Quaternion gradient
            dqdx = eigvec_grad(w, v_, ind=-1, mat_grad=dF)

            dvdq = np.zeros((3, 4))
            dvdq[:, 0] = dvdq0 * quat[1:]
            dvdq[:, 1:] = np.diag((prefac, prefac, prefac))

            # Gradient of exponential map from chain rule.
            # See bottom-left on 214108-3 in [1], after Eq. (11).
            dvdx = np.einsum("ji,ik->jk", dvdq, dqdx)

            # compare_to_geometric(c3d, ref_c3d, dR, dF, dqdx, dvdx)
            row = np.zeros_like(coords3d)
            if index is None:
                return v, dvdx.reshape(3, -1)
            row[indices] = dvdx[index].reshape(-1, 3)
            return v[index], row.flatten()
        return v[index]

    @staticmethod
    def _jacobian(coords3d, indices, index=0, ref_coords3d=None):
        """Not implemented!"""
        size = len(indices) * 3
        return np.zeros(size * size)


class RotationA(Rotation):
    index = 0


class RotationB(Rotation):
    index = 1


class RotationC(Rotation):
    index = 2



import json

import numpy as np

from pysisyphus.constants import ANG2BOHR
from pysisyphus.Geometry import Geometry
from pysisyphus.elem_data import INV_ATOMIC_NUMBERS


def parse_cjson(fn):
    with open(fn) as handle:
        cml = json.load(handle)
    atms = cml["atoms"]
    coords = np.array(atms["coords"]["3d"], dtype=float) * ANG2BOHR
    atom_nums = atms["elements"]["number"]
    atoms = tuple([INV_ATOMIC_NUMBERS[num].capitalize() for num in atom_nums])

    return atoms, coords.flatten()


def geom_from_cjson(fn, **kwargs):
    atoms, coords = parse_cjson(fn)
    geom = Geometry(atoms, coords.flatten(), **kwargs)
    return geom



from datetime import timedelta
import sys
import time

from pysisyphus.helpers_pure import highlight_text
from pysisyphus.TablePrinter import TablePrinter


def run_perf(
    geom,
    calc_getter,
    pals=None,
    mems=None,
    pal_range=None,
    mem_range=None,
    repeat=1,
    kind="forces",
):
    assert repeat > 0
    assert pals or pal_range
    assert mems or mem_range

    func_names = {
        "energy": "get_energy",
        "forces": "get_forces",
        "hessian": "get_hessian",
    }

    if pal_range is not None:
        pals = range(*pal_range)
    try:
        pal_iter = tuple(pals)
    except TypeError:  # When 'pals' is a single integer
        pal_iter = tuple((pals,))

    if mem_range is not None:
        mems = range(*mem_range)
    try:
        mem_iter = tuple(mems)
    except TypeError:  # When 'mems' is a single integer
        mem_iter = tuple((mems,))

    results = {}
    for pal in pal_iter:
        print(f"{pal=} and ", end="")
        for mem in mem_iter:
            print(f"{mem=}:")
            for i in range(repeat):
                print(f"\tRunning {kind} cycle {i:02d} ... ", end="")
                calc = calc_getter()
                calc.pal = pal
                calc.mem = mem
                start = time.time()
                func = getattr(calc, func_names[kind])
                _ = func(geom.atoms, geom.cart_coords)
                dur = time.time() - start
                td = timedelta(seconds=dur)
                print(f"finished in {td} h")
                sys.stdout.flush()
                key = (pal, mem)
                results.setdefault(key, list()).append(td)
            print()
    return results


def print_perf_results(results):
    pal_avgs = dict()
    pal_min = 1_000_000_000  # Dummy value
    for i, (key, tds) in enumerate(sorted(results.items())):
        pal, mem = key
        pal_min = min(pal, pal_min)
        print(highlight_text(f"{i:03d}: pal={pal}, mem={mem} MB"))
        for j, td in enumerate(tds):
            print(f"Run {j:03d}: {td} h")
        avg = sum(tds, timedelta(0)) / len(tds)
        # pal_avgs.setdefault(pal, list()).append(avg)
        pal_avgs[pal] = avg
        print(f"    Avg: {avg} h")
        print()

    avg_min = pal_avgs[pal_min]
    print(f"   pal_min: {pal_min} core(s)")
    print("efficiency: speedup / (pal / pal_min)\n")
    header = ("pal", "avg. / h", "speedup", "efficieny")
    col_fmts = ("int", "str", "float", "float")
    table = TablePrinter(header, col_fmts, width=20)
    table.print_header()
    for pal, avg in pal_avgs.items():
        speedup = avg_min / avg
        speedup_per_pal = speedup / (pal / pal_min)
        table.print_row((pal, str(avg), speedup, speedup_per_pal))



# [1] https://doi.org/10.1002/jcc.540140615
#     Comparison of the Boys and Pipek–Mezey localizations in the local
#     correlation approach and automatic virtual basis selection
#     Boughton, Pulay, 1993
# [2] https://doi.org/10.1063/1.2360264
#     Fast noniterative orbital localization for large molecules
#     Aquilante, Pedersen, 2006
# [3] https://doi.org/10.1063/1.3042233
#     Constructing diabatic states from adiabatic states: Extending
#     generalized Mulliken–Hush to multiple charge centers with Boys localization
#     Subotnik, Yeganeh, Cave, Ratner, 2008
# [4] https://doi.org/10.1063/1.1681683
#     Localized molecular orbitals for polyatomic molecules.
#     I. A comparison of the Edmiston‐Ruedenberg and Boys localization methods
#     Kleier, Halgren, Hall Jr., Lipscomb, 1974
# [5] https://doi.org/10.1063/1.4894472
#     Diabatization based on the dipole and quadrupole: The DQ method
#     Hoyer, Xu, Ma, Gagliardi, Truhlar, 2014
# [6] https://pubs.acs.org/doi/pdf/10.1021/acs.jctc.2c00261
#     Implementation of Occupied and Virtual Edmiston−Ruedenberg
#     Orbitals Using Cholesky Decomposed Integrals
# [7] https://doi.org/10.1063/1.1790971
#     An efficient method for calculating maxima of homogeneous
#     functions of orthogonal matrices: Applications to localized
#     occupied orbitals


from dataclasses import dataclass
from functools import singledispatch
import itertools as it
from typing import Callable, Optional
import warnings

import numpy as np
from numpy.typing import NDArray

from pysisyphus.helpers_pure import rms
from pysisyphus.linalg import matrix_power, pivoted_cholesky
from pysisyphus.wavefunction import logger, Wavefunction
from pysisyphus.wavefunction.DIIS import DIIS


PI_QUART = np.pi / 4


@singledispatch
def cholesky(C: NDArray[float]):
    """Localization via pivoted Cholesky factorization.

    See [2].

    Parameters
    ----------
    C
        Matrix of molecular orbital coefficients to be localized.
        Shape is (naos, nmos).

    Returns
    -------
    C_loc
        Localized molecular orbital coefficient matrix of shape (naos, nmos).
    """
    naos, nmos = C.shape

    # We can't use scipy's builtin Cholesky-factorization, as it does not
    # support positive semi-definite matrices.
    L, piv, _ = pivoted_cholesky(C @ C.T)
    # Restore original ordering via permutation matrix.
    P = np.zeros((naos, naos))
    P[piv, np.arange(naos)] = 1
    C_loc = P @ L[:, :nmos]
    return C_loc


@cholesky.register
def _(wf: Wavefunction):
    """Currently localizes only C_(α,occ) MOs."""
    Cao, _ = wf.C_occ
    return cholesky(Cao)


def rot_inplace(mat, rad, i, j):
    """Inplace rotation of matrix columns mat[:, i] and mat[:, j] by 'rad' radians."""
    cos = np.cos(rad)
    sin = np.sin(rad)
    i_rot = cos * mat[:, i] + sin * mat[:, j]
    j_rot = -sin * mat[:, i] + cos * mat[:, j]
    mat[:, i] = i_rot
    mat[:, j] = j_rot


@dataclass
class JacobiSweepResult:
    is_converged: bool
    cur_cycle: int
    C: NDArray[float]
    P: float


def jacobi_sweeps(
    C: NDArray[float],
    cost_func: Callable,
    ab_func: Callable,
    callback: Optional[Callable] = None,
    max_cycles: int = 100,
    dP_thresh: float = 1e-8,
) -> JacobiSweepResult:
    """Wrapper for 2x2 Jacobi-sweeps as used in localization/diabatization.

    Parameters
    ----------
    C
        MO coefficient matrix (shape naos x nmos) or rotation matrix (nstates x nstates).
    cost_func
        Function to be maximized/minimized.
    ab_func
        Function that returns A & B values, used to calculate the angle
        for the 2x2 rotation.
    callback
        Function that is called after the 2x2 rotation took place. It takes three arguments:
        (gamma, j, k).
    max_cycles
        Maximum number of macro cycles.
    dP_thresh
        Indicate convergence when change in cost function is equal or below
        this threshold.

    Returns
    -------
    C_loc
        Localized molecular orbital coefficient matrix of shape (naos, nmos).
    """

    assert max_cycles > 0
    C = C.copy()
    _, nmos = C.shape

    if callback is None:
        callback = lambda *args: None

    P_prev = 0.0

    logger.info(f"Starting Jacobi sweeps.")
    for i in range(max_cycles):
        # Loop over pairs of MO indices and do 2x2 rotations.
        for j, k in it.combinations(range(nmos), 2):
            A, B = ab_func(j, k, C)

            if (A**2 + B**2) <= 1e-12:
                continue

            gamma = np.sign(B) * np.arccos(-A / np.sqrt(A**2 + B**2)) / 4
            assert -PI_QUART <= gamma <= PI_QUART
            # 2x2 MO rotations
            rot_inplace(C, gamma, j, k)
            callback(gamma, j, k)
        # Outside of loop over orbital pairs

        # Calculate the target cost function we want to maximize/minimize.
        P = cost_func(C)
        dP = P - P_prev
        logger.info(f"{i:03d}: {P=: >12.8f} {dP=: >12.8f}")
        if is_converged := (dP <= dP_thresh):
            logger.info(f"Jacobi sweeps converged in {i+1} cycles.")
            break
        P_prev = P
    # Outside macro cycles

    result = JacobiSweepResult(
        is_converged=is_converged,
        cur_cycle=i,
        C=C,
        P=P,
    )
    return result


@singledispatch
def pipek_mezey(C, S, ao_center_map, **kwargs) -> JacobiSweepResult:
    """Pipek-Mezey localization using Mulliken population analysis.

    Python adaption of code found in orbloc.f90 of Multiwfn 3.8.
    For now, only Mulliken population and localization exponent 2 is supported.

    Parameters
    ----------
    C
        Matrix of molecular orbital coefficients to be localized.
        Shape is (naos, nmos).
    S
        Overlap matrix of shape (naos x naos).
    ao_center_map
        Mapping between atom indices and AOs, centered at the respective atom.

    Returns
    -------
    C_loc
        Localized molecular orbital coefficient matrix of shape (naos, nmos).
    """
    _, nmos = C.shape
    centers = list(ao_center_map.keys())

    SC = S @ C

    def ab_func(j, k, C):
        Q = SC * C
        A = 0.0
        B = 0.0
        # Eq. (9) in [1]
        for center in centers:
            ao_inds = ao_center_map[center]
            Qjk = (
                C[ao_inds, j] * SC[ao_inds, k] + C[ao_inds, k] * SC[ao_inds, j]
            ).sum() / 2
            Qjj = Q[ao_inds, j].sum()
            Qkk = Q[ao_inds, k].sum()
            A += (Qjk**2) - ((Qjj - Qkk) ** 2) / 4
            B += Qjk * (Qjj - Qkk)
        return A, B

    def callback(gamma, j, k):
        # The MO rotations invalidate the SC matrix product. We update it too.
        rot_inplace(SC, gamma, j, k)

    def cost_func(C):
        # We wan't to maximize P and we monitor the progress. Eq. (8) in [1].
        P = 0.0
        for l in range(nmos):
            for center in centers:
                ao_inds = ao_center_map[center]
                Q = (C[ao_inds, l][:, None] * C[:, l] * S[ao_inds, :]).sum()
                P += Q**2
        return P

    logger.info("Pipek-Mezey localization")
    return jacobi_sweeps(C, cost_func, ab_func, callback, **kwargs)


@pipek_mezey.register
def _(wf: Wavefunction) -> JacobiSweepResult:
    """Currently localizes only C_(α,occ) MOs."""
    S = wf.S
    Cao, _ = wf.C_occ
    C_chol_loc = cholesky(Cao)
    return pipek_mezey(C_chol_loc, S, wf.ao_center_map)


def get_fb_contract(moments_ints):
    def contract(mo_j, mo_k):
        return np.einsum(
            "xkl,k,l->x",
            moments_ints,
            mo_j,
            mo_k,
            optimize=["einsum_path", (0, 1), (0, 1)],
        )

    return contract


def get_fb_ab_func(moments_ints):
    contract = get_fb_contract(moments_ints)

    def ab_func(j, k, C):
        mo_j = C[:, j]
        mo_k = C[:, k]
        jrj = contract(mo_j, mo_j)
        jrk = contract(mo_j, mo_k)
        krk = contract(mo_k, mo_k)
        A = (jrk**2).sum() - ((jrj - krk) ** 2).sum() / 4  # Eq. (9) in [4]
        B = ((jrj - krk) * jrk).sum()  # Eq. (10) in [4]
        return A, B

    return ab_func


def get_fb_cost_func(moments_ints):
    def cost_func(C):
        vals = np.einsum(
            "xkl,ki,li->ix",
            moments_ints,
            C,
            C,
            optimize=["einsum_path", (0, 1), (0, 1)],
        )
        val = ((vals[:, None, :] - vals) ** 2).sum()
        return val

    return cost_func


@singledispatch
def foster_boys(
    C: NDArray[float], dip_ints: NDArray[float], **kwargs
) -> JacobiSweepResult:
    """Foster-Boys localization.

     nMO   nMO
     ___   ___
     ╲     ╲                         2
      ╲     ╲   |                   |
      ╱     ╱   | <i|R|i> - <j|R|j> |
     ╱     ╱    |                   |
     ‾‾‾   ‾‾‾
    i = 1 j = 1

    or similarily (see eq. (6) in [4] or the appendix of [5])

     nMO
     ___
     ╲                2
      ╲    |         |
      ╱    | <i|R|i> |
     ╱     |         |
     ‾‾‾
    i = 1

    Parameters
    ----------
    dip_ints
        Dipole moment integral matrices with shape (3, naos, naos).
    C
        Matrix of molecular orbital coefficients to be localized.
        Shape is (naos, nmos).
    kwargs
        Additional keyword arguments that are passed jacobi_sweeps.

    Returns
    -------
    C_loc
        Localized molecular orbital coefficient matrix of shape (naos, nmos).
    """

    ab_func = get_fb_ab_func(dip_ints)
    cost_func = get_fb_cost_func(dip_ints)
    logger.info("Foster-Boys localization")
    return jacobi_sweeps(C, cost_func, ab_func, **kwargs)


@foster_boys.register
def _(wf: Wavefunction) -> JacobiSweepResult:
    """Currently localizes only C_(α,occ) MOs."""
    Cao, _ = wf.C_occ
    dip_ints = wf.dipole_ints()
    C_chol_loc = cholesky(Cao)
    return foster_boys(C_chol_loc, dip_ints)


def edmiston_ruedenberg_cost_func(L):
    """Edmiston-Ruedenberg cost function.

    Eq. (2) in [7]."""
    return np.einsum("Jpp,Jpp->", L, L, optimize="greedy")


def edmiston_ruedenberg_grad(L):
    """Edmiston-Ruedenberg gradient.

    Eq. (5) in [7]."""
    g1 = np.einsum("Jpq,Jqq->pq", L, L, optimize="greedy")
    g2 = np.einsum("Jqp,Jpp->pq", L, L, optimize="greedy")
    return -4 * (g1 - g2)


def edmiston_ruedenberg(C, Lao, Sao, diis_cycles=5, max_cycles=100):
    """Edmiston-Ruedenberg localization using DF integrals and DIIS.

    Parameters
    ----------
    C
        MO-coefficients of shape (naos, nmos); this means MOs should
        be in columns.

    Lao
        Density fitting tensor obtained from contracting the 2e3c-matrix
        with 2c2e**-0.5 in the AO basis. Must have shape (naux, nao, nao).
        Same as in
    Sao
        Overlap matrix of shape (nao, nao) in the AO basis.
    diis_cycles
        Integer >= 0. If > 0, DIIS is employed to accelerate convergence.
    max_cycles
        Positive integer; maximum number of localization cycles.

    Returns
    -------
    Crot
        Edmiston-Ruedenberg localized orbitals.
    """
    nmos = C.shape[1]
    # We can't keep more error vectors than MOs
    if nmos < diis_cycles:
        warnings.warn(
            f"Can keep at most {nmos} DIIS error vectors, but "
            f"{diis_cycles} were requested!."
        )
    diis_cycles = min(diis_cycles, nmos)
    if diis_cycles == 1:
        diis_cycles = 0
    started_to_store_diis = False
    specs = {
        "err_vecs": np.zeros((diis_cycles, nmos**2)),
        "R_mats": np.zeros((diis_cycles, nmos, nmos)),
        "D_mats": np.zeros((diis_cycles, nmos, nmos)),
    }
    diis = DIIS(specs)

    D = np.eye(nmos)
    C0 = C.copy()
    Crot = C.copy()
    for i in range(max_cycles):
        # Transform cholesky integrals to MO basis, step (2) in [7].
        Lpq = np.einsum("Luv,up,vq->Lpq", Lao, Crot, Crot, optimize="greedy")
        f = edmiston_ruedenberg_cost_func(Lpq)
        g = edmiston_ruedenberg_grad(Lpq)
        gnorm = np.linalg.norm(g)
        grms = rms(g)
        # Construct transformation, step (3) in [7].
        # Rji = (Xj Xi Xi Xi)
        R = np.einsum("Jji,Jii->ji", Lpq, Lpq, optimize="greedy")
        # DIIS error; lack of symmetry in R.
        err = (R - R.T).flatten()
        errrms = rms(err)
        print(
            f"Cycle {i:02d} f={f:.8f}, |g|={np.linalg.norm(g):.4f}, rms(g)={grms:>8.4e} "
            f"rms(err)={errrms:>8.4e}"
        )
        if converged := grms <= 1e-4:
            print("Edmiston-Ruedenberg localization converged!")
            break

        U = R @ matrix_power(R.T @ R, -0.5)
        D = D @ U
        if diis_cycles and (started_to_store_diis or (gnorm <= 2.0e-1)):
            diis.store(
                {
                    "err_vecs": err,
                    "R_mats": R,
                    "D_mats": D,
                }
            )
            started_to_store_diis = True
        if (diis_coeffs := diis.get_coeffs()) is not None:
            D_diis = np.einsum("i,ijk->jk", diis_coeffs, diis.get("D_mats"))
            C_diis = C0 @ D_diis
            # DIIS-2 algorithm in [7]
            R_diis = np.einsum("i,ijk->jk", diis_coeffs, diis.get("R_mats"))
            S_diis = C_diis.T @ Sao @ C_diis
            S_diis_inv = matrix_power(S_diis, -1.0)
            # Generalized eta step
            V = S_diis_inv @ R_diis @ matrix_power(R_diis.T @ S_diis_inv @ R_diis, -0.5)
            # Recalculate D from DIIS results
            D = D_diis @ V
        # Update Crot, step (4) in [7]. This either uses the D matrix obtained from DIIS
        # or the one previously calculated.
        Crot = C0 @ D
    # TODO: return JacobiSweepResult?
    return Crot, f



import re

import numpy as np

from pysisyphus.elem_data import INV_ATOMIC_NUMBERS
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers_pure import file_or_str
from pysisyphus.wavefunction.shells import Shell, FCHKShells
from pysisyphus.wavefunction import Wavefunction
from pysisyphus.wavefunction.helpers import BFType


@file_or_str(".fchk")
def parse_fchk(text):
    header_re = r"\s+".join(
        (
            r"(?P<keyword>[\w\-\s/\(\)=\*]+)",
            r"(?P<kind>[HIRCL])",
            r"(?P<length>N=)?",
            r"(?P<num>[\d\-\+\.E]+)",
        )
    )
    header_re = re.compile(header_re)

    cur_keyword = None
    lines = text.strip().split("\n")
    # title1, titl2 are not used
    _, _, *lines = lines

    conv_funcs = {
        "R": lambda line: [float(_) for _ in line.split()],  # real
        "I": lambda line: [int(_) for _ in line.split()],  # integer
        "C": lambda line: [line],  # character
        "H": lambda line: [bool(_) for _ in line.split()],  # logical
        "L": lambda line: [bool(_) for _ in line.split()],  # logical
    }

    data = {}
    for line in lines:
        line = line.strip()
        if mobj := header_re.match(line):
            cur_keyword = mobj.group("keyword").strip()
            conv_func = conv_funcs[mobj.group("kind")]
            if mobj.group("length"):
                data[cur_keyword] = list()
            else:
                data[cur_keyword] = conv_func(mobj.group("num"))[0]
            continue
        data[cur_keyword].extend(conv_func(line))
    return data


@file_or_str(".fchk")
def shells_from_fchk(text):
    data = parse_fchk(text)

    def to_arr(key, dtype):
        return np.array(data[key], dtype=dtype)

    """
    These coordinates may not coincide with the basis functions?! Nonetheless,
    basis functions are usually atom-centered. But just to be sure we use the
    entries from 'Coordinates of each shell'.
    # coords3d = to_arr("Current cartesian coordinates", float).reshape(-1, 3)
    """
    atomic_nums = to_arr("Atomic numbers", int)

    centers = to_arr("Shell to atom map", int) - 1
    ang_moms = to_arr("Shell types", int)
    prims_per_shell = to_arr("Number of primitives per shell", int)

    prim_exponents = to_arr("Primitive exponents", float)
    prim_coeffs = to_arr("Contraction coefficients", float)
    try:
        p_prim_coeffs = to_arr("P(S=P) Contraction coefficients", float)
    except KeyError:
        p_prim_coeffs = np.zeros_like(prim_coeffs)
    shell_coords3d = to_arr("Coordinates of each shell", float).reshape(-1, 3)

    _shells = list()
    prim_ind = 0
    for i, (center_ind, L, prim_num) in enumerate(
        zip(centers, ang_moms, prims_per_shell)
    ):
        # 0: s, 1: p, -1: sp, 2: 6d, -2: 5d, 3: 10f, -3: 7f, etc...
        assert L <= 1, "Only spherical basis functions are supported."
        # Convert SP-shell with L = -1 to a s-shell with L = 0.
        # The p-shell is also created later.
        L, is_sp = (0, True) if (L == -1) else (L, False)

        coeffs = list()
        p_coeffs = list()
        exps = list()
        for _ in range(prim_num):
            coeffs.append(prim_coeffs[prim_ind])
            p_coeffs.append(p_prim_coeffs[prim_ind])
            exps.append(prim_exponents[prim_ind])
            prim_ind += 1
        atomic_num = atomic_nums[center_ind]
        center = shell_coords3d[i]

        def append_shell(coeffs, L=L):
            shell = Shell(
                L=abs(L),
                atomic_num=atomic_num,
                center=center,
                coeffs=coeffs,
                exps=exps,
                center_ind=center_ind,
            )
            _shells.append(shell)

        # Create shell with actual L.
        # If we deal with a SP shell we also create the P shell below.
        append_shell(coeffs)
        if is_sp:
            # Explicitly create p-shell
            append_shell(p_coeffs, L=1)
    shells = FCHKShells(_shells)
    return shells


def atoms_from_data(data):
    atomic_numbers = data["Atomic numbers"]
    atoms = tuple([INV_ATOMIC_NUMBERS[Z] for Z in atomic_numbers])
    return atoms


@file_or_str(".fchk")
def wavefunction_from_fchk(text):
    data = parse_fchk(text)

    charge = data["Charge"]
    mult = data["Multiplicity"]
    atoms = atoms_from_data(data)
    coords = data["Current cartesian coordinates"]
    unrestricted = "Beta Orbital Energies" in data

    def mos_for_spin(spin):
        mo_ens = data[f"{spin} Orbital Energies"]
        mo_num = len(mo_ens)
        mo_coeffs = np.array(data[f"{spin} MO coefficients"])
        mo_coeffs = mo_coeffs.reshape(mo_num, mo_num)
        return mo_coeffs

    C_a = mos_for_spin("Alpha").T
    occ_a = data["Number of alpha electrons"]
    if unrestricted:
        C_b = mos_for_spin("Beta")
        occ_b = data["Number of beta electrons"]
    else:
        C_b = C_a.copy()
        occ_b = occ_a
    C = np.stack((C_a, C_b))
    shells = shells_from_fchk(text)
    return Wavefunction(
        atoms=atoms,
        coords=coords,
        charge=charge,
        mult=mult,
        unrestricted=unrestricted,
        occ=(occ_a, occ_b),
        C=C,
        bf_type=BFType.PURE_SPHERICAL,
        shells=shells,
    )


@file_or_str(".fchk")
def geom_from_fchk(text, **geom_kwargs):
    data = parse_fchk(text)
    atoms = atoms_from_data(data)

    try:
        coords = data["Opt point       1 Geometries"]
    except KeyError:
        coords = data["Current cartesian coordinates"]
    coords = np.array(coords)
    coords = coords.reshape(-1, 3 * len(atoms))

    geoms = [Geometry(atoms, coords_, **geom_kwargs) for coords_ in coords]
    if len(geoms) == 1:
        geoms = geoms[0]
    return geoms



import argparse
import sys

from pysisyphus.config import p_DEFAULT, T_DEFAULT
from pysisyphus.thermo import get_thermoanalysis_from_hess_h5, print_thermoanalysis


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("hess_h5", help="HDF5 Hessian file from pysisyphus.")
    parser.add_argument("-T", default=T_DEFAULT, type=float, help="Temperature")
    parser.add_argument("-p", default=p_DEFAULT, type=float, help="Pressure")
    parser.add_argument("--pg", default="c1", help="Point group.")
    parser.add_argument(
        "--calorie",
        action="store_true",
        help="Output in kcal mol⁻¹ instead of kJ mol⁻¹.",
    )

    return parser.parse_args(args)


def run_thermo():
    args = parse_args(sys.argv[1:])
    hess_h5 = args.hess_h5
    T = args.T
    p = args.p
    point_group = args.pg
    unit = "calorie" if args.calorie else "joule"
    thermo = get_thermoanalysis_from_hess_h5(hess_h5, T=T, p=p, point_group=point_group)
    print_thermoanalysis(thermo, unit=unit)



# [1] https://doi.org/10.1016/j.cplett.2021.138970
#     The dynamical significance of valley-ridge inflection points
#     Víctor J.García-Garrido, Stephen Wiggins

import numpy as np

from pysisyphus.calculators.AnaPotBase import AnaPotBase


class VRIPot(AnaPotBase):
    def __init__(self, **kwargs):
        VTS = 0.5
        xs = 1.0
        xi = 0.3265
        A = 1.0
        B = 1.0
        C = 1.0
        lim = 1.5
        lims = (-lim, lim)
        kwargs_ = {
            "V_str": (
                f"{VTS} / {xs}**4 * x**2 * (x**2 - 2*{xs}**2)"
                f"+ {A}*y**2 * ({xi}-x) "
                f"+ y**4 * ({B} + {C}*x)"
            ),
            "xlim": lims,
            "ylim": lims,
            "levels": np.linspace(-3 * VTS, 2 * VTS, 65),
            # "minima": ((-1.05274, 1.02776, 0), (1.94101, 3.85427, 0)),
            # "saddles": ((0.6117313, 1.4929732, 0.0),),
        }
        _ = kwargs_["V_str"]
        kwargs_.update(kwargs)
        super().__init__(**kwargs_)

    def __str__(self):
        return "VRIPot calculator"
